from anndata import AnnData
from typing import Union, Optional
import pandas as pd
import numpy as np
from .distance import Distance
from .distances import Distances
from .align import DistanceAlignment
from . import logger

def harmonize(adata: AnnData,
              #dataset & cell type info
              dataset: str, cell_type: str, use_rep: Optional[str] = None, metric: Optional[str] = None,
              #PCT the distances or directly calculate
              use_pct: bool = False,
              #filter and normalize; note PCT will normalize after prediction
              filter_cells: bool = False, normalize: bool = True, Gaussian_kernel: bool = False,
              #PCT train
              F_test_prune: bool = True, p_thres: float = 0.05, random_state: int = 2,
              #order of datasets
              dataset_order: Optional[Union[list, tuple, np.ndarray, pd.Series, pd.Index]] = None, reorder_dataset: bool = True,
              #align params
              minimum_unique_percents: Union[list, tuple, np.ndarray, pd.Series, pd.Index, float] = (0.4, 0.5, 0.6, 0.7, 0.8),
              minimum_divide_percents: Union[list, tuple, np.ndarray, pd.Series, pd.Index, float] = (0.1, 0.15, 0.2),
              maximum_novel_percent: float = 0.05,
              #reannotate
              reannotate: bool = True, prefix: str = '',
              #to PCT train
              **kwargs) -> DistanceAlignment:
    """
    PCT-based cell type harmonization across datasets/batches.

    Parameters
    ----------
    adata
        An :class:`~anndata.AnnData` object containing different datasets/batches and cell types.
        In most scenarios, the format of the expression `.X` in the AnnData is flexible (normalized, log-normalized, z-scaled, etc.).
        However, when `use_rep` is specified as `'X'` (or `X_pca` is not detected in `.obsm` and no other latent representations are provided), `.X` should be log-normalized (to a constant total count per cell).
    dataset
        Column name (key) of cell metadata specifying dataset/batch information.
    cell_type
        Column name (key) of cell metadata specifying cell type information.
    use_rep
        Representation used to calculate distances. This can be `'X'` or any representations stored in `.obsm`.
        Default to the PCA coordinates if present (if not, use the expression matrix `X`).
    metric
        Metric to calculate the distance between each cell and each cell type. Can be `'euclidean'`, `'cosine'`, `'manhattan'` or any metrics applicable to :func:`sklearn.metrics.pairwise_distances`.
        Default to `'euclidean'` if latent representations are used for calculating distances, and to `'correlation'` if the expression matrix is used.
    use_pct
        Whether to use a predictive clustering tree to infer cross-dataset cell type distances.
        Setting to `True` will calculate distances based on PCT, which is intended for datasets with large batch effects.
        (Default: `False`)
    filter_cells
        Whether to filter out cells whose gene expression profiles do not correlate most with the eigen cell they belong to (i.e., correlate most with other cell types).
        Setting to `True` will speed up the run as only a subset of cells are used, but will render the remaining cells (i.e., filtered cells) unannotated (see the `reannotate` argument).
        (Default: `False`)
    normalize
        Whether to normalize the distance matrix if `use_pct = False` (or normalize the predicted distance if `use_pct = True`).
        (Default: `True`)
    Gaussian_kernel
        Whether to apply the Gaussian kernel to the distance matrix.
        (Default: `False`)
    F_test_prune
        Whether to use a F-test to prune the tree by removing unnecessary splits.
        (Default: `True`)
    p_thres
        p-value threshold for pruning nodes after F-test.
        (Default: `0.05`)
    random_state
        Random seed for feature shuffling during PCT training.
        (Default: `2`)
    dataset_order
        Order of datasets to be aligned. If this argument is specified, `reorder_dataset` is ignored.
        Default to the order in the distance matrix (alphabetical order in most cases) if `reorder_dataset = False`.
    reorder_dataset
        Whether to reorder datasets based on their pairwise similarities.
        (Default: `True`)
    minimum_unique_percents
        The minimum cell assignment fraction(s) to claim a cell type as uniquely matched to a cell type from the other dataset.
        By default, five values will be tried (0.4, 0.5, 0.6, 0.7, 0.8) to find the one that produces least alignments in each harmonization iteration.
    minimum_divide_percents
        The minimum cell assignment fraction(s) to claim a cell type as divisible into two or more cell types from the other dataset.
        By default, three values will be tried (0.1, 0.15, 0.2) to find the one that produces least alignments in each harmonization iteration.
    maximum_novel_percent
        The maximum cell assignment fraction to claim a cell type as novel to a given dataset.
        (Default: `0.05`)
    reannotate
        Whether to reannotate cells into harmonized cell types.
        (Default: `True`)
    prefix
        Column prefix for the reannotation data frame.
    **kwargs
        Other keyword arguments passed to :class:`~cellhint.pct.PredictiveClusteringTree`.

    Returns
    ----------
    DistanceAlignment
        A :class:`~cellhint.align.DistanceAlignment` object. Four important attributes within this class are:
        1) :attr:`~cellhint.align.DistanceAlignment.base_distance`, cross-dataset distances between all cells and all cell types.
        2) :attr:`~cellhint.align.DistanceAlignment.relation`, the harmonization table.
        3) :attr:`~cellhint.align.DistanceAlignment.groups`, high-hierarchy cell types categorizing rows of the harmonization table.
        4) :attr:`~cellhint.align.DistanceAlignment.reannotation`, reannotated cell types and cell type groups.
    """
    #raw counts are not allowed to build trees
    if use_pct and adata.X.min() >= 0 and float(adata.X.max()).is_integer():
        raise ValueError(
                f"🛑 `.X` of the AnnData is detected to be raw counts, which is not suitable for building PCT")
    #build PCT using all genes is not realistic
    if use_pct and adata.n_vars > 15000:
        logger.warn(f"⚠️ Warning: {adata.n_vars} features are used and may take long time for building PCT. Subsetting the AnnData into HVGs is recommended")
    #generate a combined `Distance`
    if use_pct:
        separate_distances = Distances(adata, dataset = dataset, cell_type = cell_type, use_rep = use_rep, metric = metric, n_jobs = -1)
        if filter_cells:
            separate_distances.filter_cells(check_symmetry = False)
        separate_distances.train(F_test_prune = F_test_prune, p_thres = p_thres, random_state = random_state, **kwargs)
        combined_distance = separate_distances.predict(normalize = normalize, return_distance = True, Gaussian_kernel = Gaussian_kernel)
    else:
        combined_distance = Distance.from_adata(adata, dataset = dataset, cell_type = cell_type, use_rep = use_rep, metric = metric, n_jobs = -1, check_params = True)
        if filter_cells:
            combined_distance.filter_cells(check_symmetry = False)
        if normalize:
            combined_distance.normalize(Gaussian_kernel = Gaussian_kernel, rank = True, normalize = True)
    #before cell type alignment
    combined_distance.assign()
    alignment = DistanceAlignment(combined_distance, check = False, dataset_order = dataset_order, row_normalize = True, maximum_novel_percent = maximum_novel_percent)
    if dataset_order is None and reorder_dataset:
        logger.info(f"🏆 Reordering datasets")
        alignment.reorder_dataset()
    #cell type alignment
    alignment.best_align(dataset_order = None, minimum_unique_percents = minimum_unique_percents, minimum_divide_percents = minimum_divide_percents)
    #reannotate
    if reannotate:
        logger.info(f"🖋️ Reannotating cells")
        alignment.reannotate(prefix = prefix)
    logger.info(f"✅ Harmonization done!")
    #return
    return alignment

harmonise = harmonize



import numpy as np
import pandas as pd
from typing import Union, Optional
from .symbols import NOVEL, REMAIN
import sys
try:
    import plotly.graph_objects as go
except ImportError:
    pass
import matplotlib
from matplotlib import pyplot as plt
import os
SEP1 = '_____'
SEP2 = 'CXCXCXCX'
DEFAULT_SANKEY_COLORS = ['#2E91E5', '#E15F99', '#1CA71C', '#FB0D0D', '#DA16FF', '#B68100', '#750D86', '#EB663B', '#511CFB', '#00A08B', '#FB00D1', '#FC0080', '#B2828D', '#6C7C32', '#778AAE', '#862A16',
                         '#A777F1', '#620042', '#1616A7', '#DA60CA', '#6C4516', '#0D2A63', '#AF0038', '#FD3216', '#00FE35', '#6A76FC', '#FED4C4', '#FE00CE', '#0DF9FF', '#F6F926', '#FF9616', '#479B55',
                         '#EEA6FB', '#DC587D', '#D626FF', '#6E899C', '#00B5F7', '#B68E00', '#C9FBE5', '#FF0092', '#22FFA7', '#E3EE9E', '#86CE00', '#BC7196', '#7E7DCD', '#FC6955', '#E48F72']

def _qq_order(sub_relation, ordered_cols):
    """
    Disentangle the cross connections in tree plot.
    """
    sub_relation2 = sub_relation[ordered_cols].copy()
    nms = sub_relation2.iloc[:, 0].value_counts().index.values
    for i in range(len(ordered_cols) - 1):
        i += 1
        val1 = sub_relation2.iloc[:, :i].agg('_'.join, axis = 1)
        val2 = sub_relation2.iloc[:, i]
        tb = pd.crosstab(val1, val2).loc[nms]
        nms = []
        flag = 0
        for j in range(tb.shape[0]):
            ind1 = (tb.iloc[j] > 0).values
            if j < (tb.shape[0] - 1):
                ind2 = (tb.iloc[j + 1] > 0).values
            else:
                ind2 = np.repeat([False], tb.shape[1])
            ind = ind1 & ind2
            all_nms = tb.columns[ind1][np.argsort(tb.loc[:, ind1].iloc[j])[::-1]]
            nm1 = tb.index[j]
            if ind.sum() > 0:
                last_nm = tb.columns[ind][0]
                mid_nms = np.setdiff1d(all_nms, last_nm)
                if flag == 1:
                    first_nm = fnm
                    mid_nms = np.setdiff1d(mid_nms, first_nm)
                    nms.append(nm1 + '_' + first_nm)
                for md in mid_nms:
                    nms.append(nm1 + '_' + md)
                nms.append(nm1 + '_' + last_nm)
                flag = 1
                fnm = last_nm
            else:
                if flag == 1:
                    all_nms = np.setdiff1d(all_nms, fnm)
                    nms.append(nm1 + '_' + fnm)
                for md in all_nms:
                    nms.append(nm1 + '_' + md)
                flag = 0
    sub_relation.index = sub_relation2.agg('_'.join, axis = 1)
    return sub_relation.loc[nms]

#def _qq_order2(sub_relation, ordered_cols):
#    sub_relation2 = sub_relation[ordered_cols].copy()
#    nms = sub_relation2.iloc[:, 0].value_counts().index.values
#    for i in range(len(ordered_cols)-1):
#        i += 1
#        val1 = sub_relation2.iloc[:, :i].agg('@'.join, axis=1)
#        val2 = sub_relation2.iloc[:, i]
#        tb = pd.crosstab(val1, val2)
#        tb = tb.loc[nms]
#        csum = (tb > 0).sum(0)
#        if csum.max() > 1:
#            idxs = np.argsort(csum)
#            for idx in idxs[csum[idxs] > 1]:
#                idxs2 = np.where(tb.iloc[:, idx] > 0)[0]
#                if np.any(np.diff(idxs2) > 1):
#                    ti = idxs2[0]
#                    tnm = np.array(nms[ti].split('@'))
#                    for idx2 in idxs2[1:]:
#                        nm = np.array(nms[idx2].split('@'))
#                        if idx2 == 0:
#                            nm2 = nms[idx2+1].split('_')
#                        elif idx2 + 1 < len(nms):
#                            nm2 = nms[idx2-1].split('_') + nms[idx2+1].split('_')
#                        else:
#                            nm2 = nms[idx2-1].split('_')
#                        if len(set(nm) & set(nm2)) == 0:
#                            mv_idxs.append(idx2)
#        nms = []
#        flag = 0
#        for j in range(tb.shape[0]):
#            ind1 = (tb.iloc[j] > 0).values
#            if j < (tb.shape[0] - 1):
#                ind2 = (tb.iloc[j+1] > 0).values
#            else:
#                ind2 = np.repeat([False], tb.shape[1])
#            ind = (ind1 & ind2)
#            all_nms = tb.columns[ind1][np.argsort(tb.loc[:, ind1].iloc[j])[::-1]]
#            nm1 = tb.index[j]
#            if ind.sum() > 0:
#                last_nm = tb.columns[ind][0]
#                mid_nms = np.setdiff1d(all_nms, last_nm)
#                if flag == 1:
#                    first_nm = fnm
#                    mid_nms = np.setdiff1d(mid_nms, first_nm)
#                    nms.append(nm1 + '_' + first_nm)
#                for md in mid_nms:
#                    nms.append(nm1 + '_' + md)
#                nms.append(nm1 + '_' + last_nm)
#                flag = 1
#                fnm = last_nm
#            else:
#                if flag == 1:
#                    all_nms = np.setdiff1d(all_nms, fnm)
#                    nms.append(nm1 + '_' + fnm)
#                for md in all_nms:
#                    nms.append(nm1 + '_' + md)
#                flag = 0
#    sub_relation.index = sub_relation2.agg('_'.join, axis=1)
#    sub_relation = sub_relation.loc[nms]
#    return sub_relation

def _relation_to_data(relation, return_sankey: bool = True) -> Union[pd.DataFrame, tuple]:
    """
    For internal use. Turn the harmonization result into Sankey input.
    """
    datasets = relation.columns[0::2]
    relation = relation[datasets].copy()
    relation.index = np.arange(relation.shape[0])
    #add prefix for cell types
    for i in relation.index:
        for j in datasets:
            content = relation.loc[i, j]
            if content not in [NOVEL, REMAIN]:
                relation.loc[i, j] = j + SEP1 + content
    #rename columns
    for _j in range(len(datasets)):
        j = datasets[_j]
        mapping_NOVEL = dict()
        mapping_REMAIN = dict()
        suffix_NOVEL = 0
        suffix_REMAIN = 0
        for i in relation.index:
            content = relation.loc[i, j]
            if content in [NOVEL, REMAIN]:
                celltypes = relation.loc[i].values
                relay = celltypes[~np.isin(celltypes, [NOVEL, REMAIN])][0] if _j == 0 else celltypes[_j-1]
                if content == NOVEL:
                    if relay not in mapping_NOVEL:
                        suffix_NOVEL += 1
                        mapping_NOVEL[relay] = f"{NOVEL}{SEP1}{suffix_NOVEL}"
                    relation.loc[i, j] = j + SEP1 + mapping_NOVEL[relay]
                else:
                    if relay not in mapping_REMAIN:
                        suffix_REMAIN += 1
                        mapping_REMAIN[relay] = f"{REMAIN}{SEP1}{suffix_REMAIN}"
                    relation.loc[i, j] = j + SEP1 + mapping_REMAIN[relay]
    if not return_sankey:
        return relation
    #label
    label = np.concatenate([np.unique(relation[j]) for j in datasets])
    refer = pd.Series(label).reset_index().set_index(0)
    #link = source + target + value
    link = pd.DataFrame(columns = ['source', 'target', 'value'])
    for k in range(len(datasets) - 1):
        vc = (relation[datasets[k]] + SEP2 + relation[datasets[k+1]]).value_counts().reset_index()
        count_column = vc.columns[1]
        vc[['source', 'target']] = vc['index'].str.split(SEP2, expand=True).values
        vc.rename(columns = {count_column: 'value'}, inplace = True)
        link = pd.concat([link, vc[['source', 'target', 'value']]])
    source = refer.loc[link.source.values, 'index'].values
    target = refer.loc[link.target.values, 'index'].values
    value = link.value.astype(int).values
    return relation, go.Sankey(node = dict(label = label), link = dict(source = source, target = target, value = value))

def _identify_relation_groups(relation, group_prefix: str = 'Group', order_row: bool = True, order_column: bool = False) -> tuple:
    """
    For internal use. Identify cell type groups based on the cell type harmonization result.
    """
    new_relation = _relation_to_data(relation, False)
    datasets = new_relation.columns
    dup_celltypes = np.unique(np.concatenate([new_relation[dataset].values[new_relation[dataset].duplicated()] for dataset in datasets]))
    if len(dup_celltypes) > 0:
        groups = np.full(new_relation.shape[0], f"{group_prefix}0", dtype = object)
        rownames = new_relation.index
        i = 0
        while len(rownames) > 0:
            i += 1
            receive = []
            provide = [rownames[0]]
            while len(provide) > 0:
                pp = provide.pop()
                receive.append(pp)
                for dataset in datasets:
                    celltype = new_relation.loc[pp, dataset]
                    if celltype in dup_celltypes:
                        extends = new_relation[new_relation[dataset] == celltype].index
                        provide.extend(extends.tolist())
                provide = list(set(provide).difference(receive))
            groups[new_relation.index.isin(receive)] = f"{group_prefix}{i}"
            rownames = rownames[~np.isin(rownames, receive)]
    else:
        groups = np.array([f"{group_prefix}{i+1}" for i in range(new_relation.shape[0])], dtype = object)
    ##remove in the future-->
    assert np.all(groups != f"{group_prefix}0")
    ##<<-remove in the future
    if order_row:
        df = []
        gs = []
        for j in range(1, len(np.unique(groups))+1):
            each_group = f"{group_prefix}{j}"
            sub_relation = new_relation[groups == each_group]
            if sub_relation.shape[0] > 1:
                col_uniques = sub_relation.apply(pd.Series.nunique).values
                ordered_cols = datasets[np.argsort(col_uniques)]
                #sub_relation = sub_relation.sort_values(by = ordered_cols.tolist())
                sub_relation = _qq_order(sub_relation, ordered_cols)
                if order_column:
                    sub_relation = sub_relation[ordered_cols]
                    sub_relation.columns = [f"D{x+1}" for x in range(sub_relation.shape[1])]
                    flag_unique = ~sub_relation.isin(dup_celltypes)
                    for row_index in range(flag_unique.shape[0]):
                        row_series = flag_unique.iloc[row_index]
                        col_indices = []
                        last_col = flag_unique.shape[1] - 1
                        while row_series.values[last_col]:
                            col_indices.append(last_col)
                            last_col -= 1
                        col_indices.reverse()
                        if len(col_indices) >= 2:
                            is_blank = np.isin([x.split(SEP1)[1] for x in sub_relation.iloc[row_index, col_indices]], [NOVEL, REMAIN])
                            sub_relation.iloc[row_index, col_indices] = sub_relation.iloc[row_index, col_indices].values[np.argsort(~is_blank)]
            else:
                if order_column:
                    is_blank = np.isin(relation.loc[groups == each_group, datasets].values[0], [NOVEL, REMAIN])
                    sub_relation = sub_relation[datasets[np.argsort(~is_blank)]]
                    sub_relation.columns = [f"D{x+1}" for x in range(sub_relation.shape[1])]
            df.append(sub_relation)
            gs.extend([each_group] * sub_relation.shape[0])
        return np.array(gs, dtype = object), pd.concat(df, axis = 0, ignore_index = True)
    return groups, new_relation

def _mix_colors(cols) -> str:
    """
    For internal use. Get the blended color.
    """
    return matplotlib.colors.to_hex(np.array([matplotlib.colors.to_rgb(col) for col in cols]).mean(axis = 0))

def _new_relation_to_color(new_relation, node_color, novel_node_color, remain_node_color, cmap = 'Reds') -> dict:
    """
    For internal use. Get the cell-type-to-color mapping.
    """
    map_color = {NOVEL: novel_node_color, REMAIN: remain_node_color}
    if node_color is None:
        for i in new_relation.index:
            row_color = DEFAULT_SANKEY_COLORS[i % len(DEFAULT_SANKEY_COLORS)]
            for j in new_relation.columns:
                content = new_relation.loc[i, j]
                if content.split(SEP1)[1] in [NOVEL, REMAIN]:
                    continue
                if content not in map_color:
                    map_color[content] = row_color
                else:
                    map_color[content] = _mix_colors([map_color[content], row_color])
    elif isinstance(node_color, pd.DataFrame):
        node_color = node_color.copy()
        map_values = node_color.iloc[:, 2].values
        if not isinstance(map_values[0], str):
            q10 = np.quantile(map_values, 0.10)
            q85 = np.quantile(map_values, 0.85)
            map_values[map_values <= q10] = q10
            map_values[map_values >= q85] = q85
            map_values = (map_values - map_values.min()) / map_values.ptp()
            map_values = plt.get_cmap(cmap, 256)(map_values)
            map_values = [matplotlib.colors.to_hex(map_value) for map_value in map_values]
        node_color['_combination'] = node_color.iloc[:, 0].astype(str) + SEP1 + node_color.iloc[:, 1].astype(str)
        map_color.update(dict(zip(node_color['_combination'].values, map_values)))
    else:
        raise TypeError(
                f"🛑 Please provide `node_color` as a data frame")
    return map_color

def sankeyplot(alignment,
           #node colors
           node_color: Optional[pd.DataFrame] = None, novel_node_color: str = '#FFFFFF', remain_node_color: str = '#F0F0F0',
           #link color
           link_color: Optional[str] = None,
           #figure elements
           title: str = 'CellHint label harmonization',
           #figure size
           show: bool = True, save: Union[str, bool] = False, width: Optional[int] = None, height: Optional[int] = None,
           #to fig.update_layout
           layout_dict: dict = {},
           #to fig.update_traces
           trace_dict: dict = {},
           #for developer use
           expand_label: bool = False,
          ) -> None:
    """
    Generate a Sankey diagram showing the CellHint label harmonization in a qualitative manner.

    Parameters
    ----------
    alignment
        A :class:`~cellhint.align.DistanceAlignment` or :class:`~pandas.DataFrame` object representing the harmonization result.
    node_color
        A :class:`~pandas.DataFrame` with three consecutive columns representing dataset, cell type, and color, respectively.
        Default to a color scheme that allows matched cell types to have the same colors.
    novel_node_color
        Color of dataset-specific (i.e., novel) cell types.
        (Default: `'#FFFFFF'`)
    remain_node_color
        Color of remaining unresolved cell types.
        (Default: `'#F0F0F0'`)
    link_color
        Color of links. Default to translucent grey as used in plotly.
    title
        Figure title.
        (Default: `'CellHint label harmonization'`)
    show
        Whether to show the plot.
        (Default: `True`)
    save
        Whether to save the plot. This can also be a figure filename.
        Supported figure suffixes are: .html, .png, .jpg, .jpeg, .webp, .svg, .pdf, .eps.
        (Default: `False`)
    width
        Figure width in pixels.
        Default to 700 in a canonical Plotly setting.
    height
        Figure height in pixels.
        Default to 450 in a canonical Plotly setting.
    layout_dict
        A dict passed to the method `.update_layout` of :class:`plotly.graph_objects.Figure` for setting the figure layout.
        Example keys include `plot_bgcolor` which sets the plot area color, `font_color` which sets the text color, `font_size` which sets the text size, etc.
    trace_dict
        A dict passed to the method `.update_traces` of :class:`plotly.graph_objects.Figure` for setting the Sankey plot.
        Example keys include `note_pad` which sets the padding between nodes, `link_line_color` which sets the link border color, `orientation` which sets the plot orientation, etc.
    expand_label
        Ignored. Whether to show the unique expanded labels. Only for developer use.
        (Default: `False`)

    Returns
    ----------
    None
    """
    if 'plotly' not in sys.modules:
        logger.warn(f"⚠️ Warning: to draw a Sankey diagram, package `plotly` is required. Please install `plotly` first")
        return
    if isinstance(alignment, pd.DataFrame):
        relation = alignment
    elif hasattr(alignment, 'relation'):
        relation = alignment.relation
    else:
        raise TypeError(
                f"🛑 Please provide correct input - either a DistanceAlignment or a data frame")
    trace = _relation_to_data(relation, return_sankey = True)[1]
    #relation2new_relation is run twice actually, but time cost is negligible; this new relation is row ordered
    new_relation = _identify_relation_groups(relation, group_prefix = 'Group', order_row = True, order_column = False)[1]
    expanded_label = trace.node.label
    original_label = pd.Series(expanded_label).str.split(SEP1, expand = True)[1].values
    blank_flag = np.isin(original_label, [NOVEL, REMAIN])
    #node color
    if isinstance(node_color, pd.DataFrame) and not np.array_equal(np.sort(node_color.iloc[:, 0].astype(str) + SEP1 + node_color.iloc[:, 1].astype(str)), np.sort(expanded_label[~blank_flag])):
        raise ValueError(
                f"🛑 Please provide a comprehensive combination of datasets and cell types in `node_color`")
    color_mapping = _new_relation_to_color(new_relation, node_color, novel_node_color, remain_node_color)
    node_color = np.array([color_mapping[x] for x in np.where(blank_flag, original_label, expanded_label)], dtype = object)
    #annotations
    datasets = new_relation.columns
    if (len(trace_dict) >= 1) and ('orientation' in trace_dict) and (trace_dict['orientation'] == 'v'):
        annotations = [dict(text = datasets[i], y = 1 - i/(len(datasets)-1), x = 0, xanchor = 'right', xref = 'paper', yanchor = 'top' if i == 0 else ('bottom' if i == len(datasets)-1 else 'middle'), yref = 'paper', showarrow = False) for i in range(len(datasets))]
    else:
        annotations = [dict(text = datasets[i], x = i/(len(datasets)-1), y = 0, yanchor = 'top', yref = 'paper', xanchor = 'left' if i == 0 else ('right' if i == len(datasets)-1 else 'center'), xref = 'paper', showarrow = False) for i in range(len(datasets))]
    #update trace and layout
    fig = go.Figure(trace)
    fig.update_traces(node_label = np.where(blank_flag, '', original_label) if not expand_label else expanded_label, node_color = node_color, link_color = link_color, **trace_dict)
    fig.update_layout(title_text = title, width = width, height = height, annotations = annotations, **layout_dict)
    if show:
        fig.show()
    if save:
        ext = os.path.splitext(save)[1] if isinstance(save, str) else '.html'
        if ext not in ['.html', '.png', '.jpg', '.jpeg', '.webp', '.svg', '.pdf', '.eps']:
            raise ValueError(
                    f"🛑 Please provide valid figure suffix: .html, .png, .jpg, .jpeg, .webp, .svg, .pdf, .eps")
        if ext == '.html':
            fig.write_html(save) if isinstance(save, str) else fig.write_html('CellHint_sankeyplot.html')
        else:
            fig.write_image(save)

def treeplot(alignment, group_celltype: bool = True, order_dataset: bool = False,
        #link
        link_color: str = '#0000007B', link_width: Optional[float] = None,
        #root and node
        node_shape: Union[list, str] = 'o', node_color: Optional[pd.DataFrame] = None, cmap: Union[matplotlib.colors.Colormap, str] = 'Reds', node_size: Optional[float] = None,
        #label
        show_label: bool = True, label_color: str = '#000000', label_size: Optional[Union[float, str]] = None, label_ha: str = 'center', label_va: str = 'top',
        #figure elements
        title: str = 'CellHint label harmonization tree',
        #show and/or save figure
        ax: Optional[matplotlib.axes.Axes] = None, figsize: Optional[Union[list, tuple]] = None, show: bool = True, save: Union[str, bool] = False,
        #link setting
        link_dict: dict = {},
        #node and root setting
        node_dict: dict = {},
        #label setting
        label_dict: dict = {},
        #for developer use
        expand_label: bool = False,
        ) -> None:
    """
    Generate a tree showing the CellHint label harmonization in a qualitative manner.

    Parameters
    ----------
    alignment
        A :class:`~cellhint.align.DistanceAlignment` or :class:`~pandas.DataFrame` object representing the harmonization result.
    group_celltype
        Whether to group cell types (rows) in the harmonization table for plotting.
        (N.B. Do not change the default value of this argument unless you know what you are doing.)
        (Default: `True`)
    order_dataset
        Whether to change the dataset order in each cell type group to manifest as hierarchy (tree).
        (Default: `False`)
    link_color
        Color of links/branches.
        (Default: `'#0000007B'`)
    link_width
        Width of links/branches in points.
        Default to 1.5 in a canonical Matplotlib setting.
    node_shape
        Shape of the node. This can also be a list of symbols for datasets that are aligned.
        (Default: `'o'`)
    node_color
        A :class:`~pandas.DataFrame` with three consecutive columns representing dataset, cell type, and color, respectively.
        Default to a color scheme that allows matched cell types to have the same colors.
        This can also be a data frame with columns of dataset, cell type, and numeric value (for mapping color gradient).
    cmap
        Color map to use. This parameter is only relevant if `node_color` is a value-mapping data frame.
        (Default: `'Reds'`)
    node_size
        Size of nodes (cell types) in points.
        Default to 6.0 in a canonical Matplotlib setting.
    show_label
        Whether to label each node with its cell type name.
        (Default: `True`)
    label_color
        Color of cell type labels.
        (Default: `'#000000'`)
    label_size
        Size of cell type labels.
        Default to 10.0 in a canonical Matplotlib setting.
    label_ha
        Horizontal alignment of cell type labels relative to the nodes.
        (Default: `'center'`)
    label_va
        Vertical alignment of cell type labels relative to the nodes.
        (Default: `'top'`)
    title
        Figure title.
        (Default: `'CellHint label harmonization tree'`)
    ax
        An :class:`~matplotlib.axes.Axes` where the tree will be drawn. Default to draw the tree on a new axes.
    figsize
        Tuple of figure width and height in inches.
        Default to auto-adjusting the figure size based on the numbers of datasets and cell types.
    show
        Whether to show the plot.
        (Default: `True`)
    save
        Whether to save the plot. This can also be a figure filename.
        (Default: `False`)
    link_dict
        A dict passed to :class:`~matplotlib.lines.Line2D` for setting the links/branches.
    node_dict
        A dict passed to :class:`~matplotlib.lines.Line2D` for setting the nodes.
    label_dict
        A dict passed to :class:`~matplotlib.text.Text` for setting cell type labels.
    expand_label
        Ignored. Whether to show the unique expanded labels. Only for developer use.
        (Default: `False`)

    Returns
    ----------
    None
    """
    if isinstance(alignment, pd.DataFrame):
        relation = alignment
    elif hasattr(alignment, 'relation'):
        relation = alignment.relation
    else:
        raise TypeError(
                f"🛑 Please provide correct input - either a DistanceAlignment or a data frame")
    new_relation = _identify_relation_groups(relation, group_prefix = 'Group', order_row = group_celltype, order_column = order_dataset)[1]
    n_col = new_relation.shape[1]
    n_row = new_relation.shape[0]
    #node coordinates
    node_coord = {}
    for col_rank in range(1, n_col + 1):
        col_content = new_relation.iloc[:, col_rank - 1]
        for celltype in np.unique(col_content):
            rows = n_row - np.where(col_content == celltype)[0]
            node_coord[celltype] = [col_rank, rows.mean()]
    #link pairs
    link_pairs = np.row_stack([new_relation.iloc[:, [i, i+1]].drop_duplicates().values for i in range(n_col - 1)])
    #ax
    if ax is None:
        figsize = [3.5*n_col, n_row/3.5] if figsize is None else figsize
        _, ax = plt.subplots(figsize = figsize)
    #links
    for start, end in link_pairs:
        xs = [node_coord[start][0], node_coord[end][0]]
        ys = [node_coord[start][1], node_coord[end][1]]
        ax.plot(xs, ys, ls = '-', marker = 'None', color = link_color, lw = link_width, **link_dict)
    if order_dataset:
        for first_cell_type in np.unique(new_relation.iloc[:, 0]):
            ax.plot([0.5, 1], [(n_row + 1)/2, node_coord[first_cell_type][1]], ls = '-', marker = 'None', color = link_color, lw = link_width, **link_dict)
    #nodes and labels
    expanded_labels = np.array(list(node_coord.keys()))
    original_labels = pd.Series(expanded_labels).str.split(SEP1, expand = True)[1].values
    blank_flag = np.isin(original_labels, [NOVEL, REMAIN])
    if isinstance(node_color, pd.DataFrame) and not np.array_equal(np.sort(node_color.iloc[:, 0].astype(str) + SEP1 + node_color.iloc[:, 1].astype(str)), np.sort(expanded_labels[~blank_flag])):
        raise ValueError(
                f"🛑 Please provide a comprehensive combination of datasets and cell types in `node_color`")
    color_mapping = _new_relation_to_color(new_relation, node_color, None, None, cmap)
    if not isinstance(node_shape, str) and len(node_shape) < n_col:
        raise ValueError(
                f"🛑 Please provide `node_shape` of length {n_col}")
    node_shapes = dict(zip(relation.columns[::2], [node_shape]*n_col)) if isinstance(node_shape, str) else dict(zip(relation.columns[::2], node_shape))
    for node, coord in node_coord.items():
        dst = node.split(SEP1)[0]
        original_label = node.split(SEP1)[1]
        if original_label in [NOVEL, REMAIN]:
            continue
        ax.plot(coord[0], coord[1], marker = node_shapes[dst], ms = node_size, color = color_mapping[node], ls = 'None', **node_dict)
        if show_label:
            ax.text(coord[0], coord[1], node if expand_label else original_label, color = label_color, size = label_size, ha = label_ha, va = label_va, **label_dict)
    if order_dataset:
        ax.plot(0.5, (n_row + 1)/2, marker = 'o', ms = node_size, color = '#000000', ls = 'None', **node_dict)
    #others
    ax.set(xlim = [0, n_col+1], ylim = [0, n_row+1], title = title)
    ax.set_axis_off()
    if not order_dataset:
        for col_rank in range(1, n_col + 1):
            ax.text(col_rank, n_row+0.5, new_relation.columns[col_rank-1], color = label_color, size = label_size, ha = 'center', va = 'center', weight = 'bold')
    #show and save
    if save:
        plt.savefig(save) if isinstance(save, str) else plt.savefig('CellHint_treeplot.pdf')
    if show:
        plt.show()
    if save:
        plt.close()

def heatmap(alignment, plot_type: str = 'similarity',
        #colors
        dataset_color: Optional[pd.DataFrame] = None, celltype_color: Optional[pd.DataFrame] = None,
        #cell
        vmin: Optional[float] = None, vmax: Optional[float] = None, cmap: str = 'RdBu_r',
        #labels
        dataset_celltype_sep: str = ': ',
        #dendrogram
        cluster: bool = True, show_row_dendrogram: bool = True, show_col_dendrogram: bool = True,
        #figure
        figsize: Union[list, tuple] = (10, 10), ax: Optional[matplotlib.axes.Axes] = None, show: bool = True, save: Union[str, bool] = False,
        #others
        **kwargs) -> None:
    """
    Generate a heatmap showing the cell type relationships within and across datasets (i.e., meta-analysis).

    Parameters
    ----------
    alignment
        A :class:`~cellhint.align.DistanceAlignment` object containing the meta-analysis result.
    plot_type
        The type of heatmap to show, being either cross-dataset cell type transcriptome similarities (`'similarity'`) or membership (`'membership'`).
        (Default: `'similarity'`)
    dataset_color
        A :class:`~pandas.DataFrame` with two consecutive columns representing dataset and color, respectively.
        Default to a CellHint color cycle.
    celltype_color
        A :class:`~pandas.DataFrame` with three consecutive columns representing dataset, cell type, and color, respectively.
        Default to a color scheme that allows matched cell types to have the same colors.
    vmin
        Minimal value to anchor the colormap.
        Default to 0 unless `plot_type = 'similarity'` and normalization is not performed during cell type harmonization.
    vmax
        Maximal value to anchor the colormap.
        Default to 1 unless `plot_type = 'similarity'` and normalization is not performed during cell type harmonization.
    cmap
        Mapping from data values to color space.
        (Default: `'RdBu_r'`)
    dataset_celltype_sep
        Separator to connect names of data sets and cell types for displaying.
    cluster
        Whether to cluster the rows and columns of the heatmap.
        (Default: `True`)
    show_{row,col}_dendrogram
        Whether to show the row/column dendrogram.
        (Default: `True`)
    figsize
        Tuple of figure width and height in inches.
        Default to 10 inches in both dimensions.
    ax
        An :class:`~matplotlib.axes.Axes` where the heatmap will be drawn. Default to draw the tree on a new axes.
    show
        Whether to show the plot.
        (Default: `True`)
    save
        Whether to save the plot. This can also be a figure filename.
        (Default: `False`)
    others
        All other parameters are the same as :func:`seaborn.clustermap` with selected tags and customized defaults.

    Returns
    ----------
    None
    """
    pass



import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)

set_level = logger.setLevel
info = logger.info
warn = logger.warning
error = logger.error
debug = logger.debug



from sklearn.tree._tree import TREE_LEAF, TREE_UNDEFINED
from sklearn.tree import DecisionTreeRegressor
from sklearn import __version__ as skv
import numpy as np
from scipy.stats import f
from scipy.sparse import spmatrix
from typing import Optional, Union
from . import logger

class PredictiveClusteringTree(DecisionTreeRegressor):
    """
    Class that uses predictive clustering trees (PCT) for multi-output prediction.
    Note this is a specialized PCT with the prototype as the mean vector and the distance measure as the (sum of) intra-cluster variance.
    Such a PCT is equivalent to CART regression tree (with specialized parameters and post-pruning).

    Parameters
    ----------
    max_depth
        Maximum possible depth of the tree, starting from the root node which has a depth of 0.
        Default to no limit.
    min_samples_split
        The minimum sample size (in absolute number or fraction) of a possible node.
        (Default: `20`)
    min_samples_leaf
        The minimum sample size (in absolute number or fraction) of a possible leaf.
        (Default: `10`)
    min_weight_fraction_leaf
        The minimum fraction out of total sample weights for a possible leaf.
        (Default: `0.0`)
    random_state
        Random seed for column (feature) shuffling before selecting the best feature and threshold.
    max_leaf_nodes
        The maximum number of leaves, achieved by keeping high-quality (i.e., high impurity reduction) nodes.
        Default to no limit.
    F_test_prune
        Whether to use a F-test to prune the tree by removing unnecessary splits.
        (Default: `True`)
    p_thres
        p-value threshold for pruning nodes after F-test.
        (Default: `0.05`)

    Attributes
    ----------
    n_features_in_
        Number of features.
    n_outputs_
        Number of outputs.
    tree_
        A :class:`~sklearn.tree._tree.Tree` object structured by parallel arrays.
    p_value
        F-test-based p-value for each node or leaf in the tree.
    """
    def __init__(self,
                 *,
                 max_depth: Optional[int] = None,
                 min_samples_split: Union[int, float] = 20,
                 min_samples_leaf: Union[int, float] = 10,
                 min_weight_fraction_leaf: float = 0.0,
                 random_state: Optional[int] = None,
                 max_leaf_nodes: Optional[int] = None,
                 F_test_prune: bool = True,
                 p_thres: float = 0.05):
        criterion = 'mse' if skv.split('.')[0] == '0' else 'squared_error'
        super().__init__(criterion = criterion,                             #L2 norm multi-output decision tree = pct with mean as prototype and SSe as distance measure
                         splitter = 'best',
                         max_depth = max_depth,
                         min_samples_split = min_samples_split,             #min_samples_leaf*2
                         min_samples_leaf = min_samples_leaf,               #minimum cell cluster size
                         min_weight_fraction_leaf = min_weight_fraction_leaf,
                         max_features = None,                               #use all features
                         random_state = random_state,
                         max_leaf_nodes = max_leaf_nodes,
                         min_impurity_decrease = 0.0,                       #split as much as possible -> no good way to select a proper value
                         ccp_alpha = 0.0)                                   #use F test for pruning, ignore ccp pruning
        self.F_test_prune = F_test_prune
        self.p_thres = p_thres

    def fit(self, X, y, sample_weight = None) -> None:
        """
        Fit a PCT with the training dataset.

        Parameters
        ----------
        X
            Sample-by-feature array-like matrix.
        y
            Sample-by-output array-like matrix.
        sample_weight
            Sample weights. Default to equal sample weights.

        Returns
        -------
        None
            Fitted and (possibly) pruned tree.
        """
        if isinstance(X, spmatrix) and ((X.indices.dtype == 'int64') or (X.indptr.dtype == 'int64')):
            super().fit(X.toarray(), y, sample_weight = sample_weight, check_input = True)
        else:
            super().fit(X, y, sample_weight = sample_weight, check_input = True)
        if self.F_test_prune:
            self.F_test()
            self.prune_tree(p_thres = self.p_thres)

    def is_node(self, index: int) -> bool:
        """
        Check whether a given index is a node.

        Parameters
        ----------
        index
            Index of the node/leaf in the arrays of the tree structure.

        Returns
        -------
        bool
            `True` or `False` indicating whether the given index is an internal node.
        """
        return self.tree_.children_left[index] != self.tree_.children_right[index]

    def F_test(self) -> None:
        """
        F test for each internal node.
        For each node, the corresponding F distribution has the degrees of freedom `n_output * (n_sample - 1)` and `n_output * (n_sample - 2)`, and the value (q) of `node_impurity * n_sample / (n_sample - 1)` divided by `(left_child_impurity * left_n_sample + right_child_impurity * right_n_sample) / (n_sample - 2)`.

        Returns
        -------
        None
            Modified tree with F-test p-values. Leaves are assigned 1 constantly.
        """
        ps = np.ones(self.tree_.node_count)
        for x in range(self.tree_.node_count):
            if self.is_node(x):
                f_score_numerator = self.tree_.impurity[x] * self.tree_.n_node_samples[x] / (self.tree_.n_node_samples[x] - 1)
                f_score_denominator = (self.tree_.impurity[self.tree_.children_left[x]]*self.tree_.n_node_samples[self.tree_.children_left[x]] +
                                       self.tree_.impurity[self.tree_.children_right[x]]*self.tree_.n_node_samples[self.tree_.children_right[x]]
                                      ) / (self.tree_.n_node_samples[x] - 2)
                f_score = f_score_numerator / f_score_denominator
                p = 1 - f.cdf(f_score, self.n_outputs_ * (self.tree_.n_node_samples[x] - 1), self.n_outputs_ * (self.tree_.n_node_samples[x] - 2))
                ps[x] = p
        self.p_value = ps

    def prune_node(self, index: int) -> None:
        """
        Prune all descendents of a given node. This node will become a leaf.

        Parameters
        ----------
        index
            Index of the node/leaf in the tree structure.

        Returns
        -------
        None
            Modified tree with all descendents of a given node pruned.
        """
        if self.is_node(index):
            stack = [index]
            while len(stack) > 0:
                 index_p = stack.pop()
                 if self.is_node(index_p):
                     stack.append(self.tree_.children_left[index_p])
                     stack.append(self.tree_.children_right[index_p])
                     self.tree_.children_left[index_p] = TREE_LEAF
                     self.tree_.children_right[index_p] = TREE_LEAF
                     self.tree_.feature[index_p] = TREE_UNDEFINED
                     self.tree_.threshold[index_p] = TREE_UNDEFINED

    def prune_tree(self, p_thres: float = 0.05) -> None:
        """
        Prune a tree based on F-test p values.

        Parameters
        ----------
        p_thres
            p-value threshold to prune nodes.
            (Default: `0.05`)

        Returns
        -------
        None
            Modified tree with unnecessary splits removed.
        """
        if p_thres < 0 or p_thres > 1:
            raise ValueError(
                    f"🛑 Please provide the `p_thres` between 0 and 1")
        stack = [0]
        while len(stack) > 0:
            index_n = stack.pop()
            if self.is_node(index_n):
                if self.p_value[index_n] < p_thres:
                    stack.append(self.tree_.children_left[index_n])
                    stack.append(self.tree_.children_right[index_n])
                else:
                    self.prune_node(index_n)

    def score(self, X, y, sample_weight = None) -> float:
        """
        Calculate the coefficient of determination between the prediction and truth.
        Different from multi-output problem where each output is calculated separately and the final R2 score is averaged across outputs, the score here is defined by considering each sample vector as a 'real' sample.

        Parameters
        ----------
        X
            Sample-by-feature query matrix.
        y
            Sample-by-output truth matrix.
        sample_weight
            Sample weights applied to squared distance of each sample.

        Returns
        ----------
        float
            Coefficient of determination.
        """
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                    f"🛑 Please provide `X` with {self.n_features_in_} columns")
        if not isinstance(y, np.ndarray):
            raise TypeError(
                    f"🛑 Please provide `y` as a numpy 2D array")
        if (X.shape[0] != y.shape[0]) or (self.n_outputs_ != y.shape[1]):
            raise ValueError(
                    f"🛑 Please provide `y` with {X.shape[0]} rows and {self.n_outputs_} columns")
        y_pred = self.predict(X)
        if sample_weight is not None:
            sample_weight = np.array(sample_weight)
            weight = sample_weight[:, np.newaxis]
        else:
            weight = 1.0
        numerator = (weight * (y - y_pred) ** 2).sum()
        denominator = (weight * (y - np.average(y, axis=0, weights = sample_weight)) ** 2).sum()
        if numerator == 0:
            return 1.0
        elif denominator == 0:
            return 0.0
        else:
            return 1 - numerator / denominator



import numpy as np
import pandas as pd
from anndata import AnnData
from typing import Optional, Union
import types
import sys
try:
    from sklearn.neighbors import DistanceMetric
except ImportError:
    from sklearn.metrics import DistanceMetric
from sklearn.neighbors import KDTree
from sklearn.metrics import pairwise_distances
from . import logger
from .symbols import UNASSIGN
from umap.umap_ import fuzzy_simplicial_set
import pynndescent
from annoy import AnnoyIndex
from scipy.spatial import cKDTree
from scipy.sparse import coo_matrix
try:
    import faiss
except ImportError:
    pass

def _locate_meta_neighbors(pca, batch_list, celltype_list, n_meta_neighbors):
    """
    Find the cell type neighbors for a given cell type (union across batches).
    """
    celltypes = np.unique(celltype_list)
    if n_meta_neighbors == 1 or len(celltypes) == 1:
        cns = {celltype: np.array([celltype], dtype = object) for celltype in celltypes}
    else:
        c2b = {celltype: np.unique(batch_list[celltype_list == celltype]) for celltype in celltypes}
        b2d = {}
        for batch in np.unique(batch_list):
            batch_flag = batch_list == batch
            sub_celltypes = np.unique(celltype_list[batch_flag])
            dists = pairwise_distances(np.array([pca[batch_flag & (celltype_list == sub_celltype)].mean(axis=0) for sub_celltype in sub_celltypes]))
            b2d[batch] = pd.DataFrame(dists, index = sub_celltypes, columns = sub_celltypes)
        cns = {celltype: np.unique(np.concatenate([b2d[b].loc[celltype].sort_values().index[:n_meta_neighbors] for b in c2b[celltype]])) for celltype in celltypes}
    if UNASSIGN in celltypes:
        cns_keys = list(cns.keys())
        for cns_key in cns_keys:
            if cns_key == UNASSIGN:
                cns[cns_key] = celltypes
            else:
                cns[cns_key] = np.unique(np.append(cns[cns_key], UNASSIGN))
    return cns

def _create_tree(pca, computation, metric, annoy_n_trees, pynndescent_n_neighbors, pynndescent_random_state):
    """
    Copied from BBKNN. Create a faiss/cKDTree/KDTree/annoy/pynndescent index for nearest neighbor lookup.
    """
    if computation == 'annoy':
        ckd = AnnoyIndex(pca.shape[1], metric = metric)
        for i in np.arange(pca.shape[0]):
            ckd.add_item(i, pca[i, :])
        ckd.build(annoy_n_trees)
    elif computation == 'pynndescent':
        ckd = pynndescent.NNDescent(pca, metric = metric, n_jobs = -1, n_neighbors = pynndescent_n_neighbors, random_state = pynndescent_random_state)
        ckd.prepare()
    elif computation == 'faiss':
        ckd = faiss.IndexFlatL2(pca.shape[1])
        ckd.add(pca)
    elif computation == 'cKDTree':
        ckd = cKDTree(pca)
    elif computation == 'KDTree':
        ckd = KDTree(pca, metric = metric)
    return ckd

def _query_tree(pca, computation, ckd, n_neighbors):
    """
    Copied from BBKNN. Query the faiss/cKDTree/KDTree/annoy/pynndescent index.
    """
    if computation == 'annoy':
        ckdo_ind = []
        ckdo_dist = []
        for i in np.arange(pca.shape[0]):
            holder = ckd.get_nns_by_vector(pca[i, :], n_neighbors, include_distances = True)
            ckdo_ind.append(holder[0])
            ckdo_dist.append(holder[1])
        ckdout = (np.asarray(ckdo_dist), np.asarray(ckdo_ind))
    elif computation == 'pynndescent':
        ckdout = ckd.query(pca, k = n_neighbors)
        ckdout = (ckdout[1], ckdout[0])
    elif computation == 'faiss':
        D, I = ckd.search(pca, n_neighbors)
        D[D < 0] = 0
        ckdout = (np.sqrt(D), I)
    elif computation == 'cKDTree':
        ckdout = ckd.query(x = pca, k = n_neighbors, workers = -1)
    elif computation == 'KDTree':
        ckdout = ckd.query(pca, k = n_neighbors)
    return ckdout

def _get_graph(pca, batch_list, celltype_list, computation, n_neighbors, n_meta_neighbors, metric, annoy_n_trees, pynndescent_n_neighbors, pynndescent_random_state, random_state):
    """
    Identify the cell-type-controlled KNN structure to be used in graph construction.
    """
    celltype_groups = _locate_meta_neighbors(pca, batch_list, celltype_list, n_meta_neighbors)
    if computation == 'faiss':
        pca = pca.astype('float32')
    knn_dists = np.zeros((pca.shape[0], n_neighbors))
    knn_indices = np.copy(knn_dists).astype(int)
    #main
    celltypes = np.unique(celltype_list)
    for celltype in celltypes:
        flag_celltype = celltype_list == celltype
        ind_celltype = np.arange(len(batch_list))[flag_celltype]
        flag_group = np.isin(celltype_list, celltype_groups[celltype])
        batches = np.unique(batch_list[flag_group])
        q, mod = divmod(n_neighbors, len(batches))
        np.random.seed(random_state)
        n_neighbors_across = np.random.permutation([q]*(len(batches)-mod) + [q+1]*mod)
        cum_sum = np.cumsum(n_neighbors_across)
        for rank, batch, n_neighbors_each in zip(range(len(batches)), batches, n_neighbors_across):
            flag_batch = batch_list == batch
            flag = flag_group & flag_batch
            if n_neighbors_each > flag.sum():
                flag = flag_batch
            ind = np.arange(len(batch_list))[flag]
            ckd = _create_tree(pca[flag], computation, metric, annoy_n_trees, pynndescent_n_neighbors, pynndescent_random_state)
            ckdout = _query_tree(pca[flag_celltype], computation, ckd, n_neighbors_each)
            col_range = np.arange(0 if rank == 0 else cum_sum[rank-1], cum_sum[rank])
            knn_indices[ind_celltype[:, np.newaxis], col_range] = ind[ckdout[1]]
            knn_dists[ind_celltype[:, np.newaxis], col_range] = ckdout[0]
    return knn_dists, knn_indices

def integrate(
          #input adata
          adata: AnnData, batch: str, cell_type: Optional[str] = None, use_rep: Optional[str] = None, n_latent: int = 50,
          #neighbors global setting
          n_neighbors: Optional[int] = None, n_meta_neighbors: int = 3, approx: bool = True, metric: Union[str, types.FunctionType, DistanceMetric] = 'euclidean',
          #if approx = True, annoy or pyNNDescent
          use_annoy: bool = True, annoy_n_trees: int = 10, pynndescent_n_neighbors: int = 30, pynndescent_random_state: int = 0,
          #if approx = False
          use_faiss: bool = True,
          #connectivities
          set_op_mix_ratio: float = 1.0, local_connectivity: int = 1, trim: Optional[int] = None,
          #random and copy
          neighbor_random_state: int = 0, copy: bool = False) -> Union[AnnData, None]:
    """
    Cell type controlled k nearest neighbors. This is a variant of BBKNN by searching neighbors across matched cell groups in different batches.
    For a given cell belonging to cell type 'c', first determine the batches that contain 'c' and its neighboring cell types, and then in each batch, search nearest neighbors out of them.

    Parameters
    ----------
    adata
        An :class:`~anndata.AnnData` object containing batch and cell type information in `.obs`, as well as latent space (e.g., `'X_pca'`) in `.obsm`.
    batch
        Column name (key) of cell metadata specifying batch information.
    cell_type
        Column name (key) of cell metadata specifying cell type information.
        Default to no cell type information provided (i.e., searching nearest neighbors in the entire batch space).
    use_rep
        Representation used to calculate distances. This can be any representations stored in `.obsm`.
        Default to the PCA coordinates (`'X_pca'`) if present.
    n_latent
        Number of latent representations used.
        Default to min(50, number of available latent representations).
    n_neighbors
        Total number of nearest neighbors for each cell. This number will be contributed equally from batches that qualify.
        Default to max(15, n) where n is the number of batches times three, meaning that each qualified batch will provide at least 3 neighbors.
        For example, if one cell type exists exclusively in one batch, then this batch needs to provide 15 neighbors.
    n_meta_neighbors
        Total number of nearest meta neighbors for each cell type in each batch (calculated from cell centroids).
        The final nearest meta neighbors are the union across batches that contain this given cell type.
        The smaller this value, the stronger bonding of the same cell type.
        Setting to 1 will make each cell search nearest neighbors only in the cell type it belongs to (i.e., forcibly clustering the same cell types).
        (Default: `3`)
    approx
        Whether to use fast approximate neighbor finding (annoy or pyNNDescent).
        (Default: `True`)
    metric
        Distance metric to use.
        (Default: `'euclidean'`)
    use_annoy
        Whether to use annoy for neighbor finding when `approx = True`. Setting `use_annoy = False` will use pyNNDescent instead.
        (Default: `True`)
    annoy_n_trees
        Number of trees to construct in the annoy forest when `approx = True` and `use_annoy = True`.
        (Default: `10`)
    pynndescent_n_neighbors
        Number of neighbors to include in the approximate neighbor graph when `approx = True` and `use_annoy = False`.
        (Default: `30`)
    pynndescent_random_state
        Random seed to use in pyNNDescent when `approx = True` and `use_annoy = False`.
        (Default: `0`)
    use_faiss
        Whether to use the faiss package to compute nearest neighbors if installed when `approx = False` and `metric = 'euclidean'`.
        (Default: `True`)
    set_op_mix_ratio
        Float between 0 and 1 controlling the blend between a connectivity matrix formed exclusively from mutual nearest neighbor pairs (0)
        and a union of all observed neighbor relationships with the mutual pairs emphasized (1).
        (Default: `1.0`)
    local_connectivity
        UMAP connectivity computation parameter controlling how many nearest neighbors of each cell are assumed to be fully connected (with a connectivity value of 1).
        (Default: `1`)
    trim
        Trim each cell to top `trim` connectivities. May help with population independence and improve the tidiness of clustering.
        Default to n_neighbors*10. Set to 0 to skip trimming.
    neighbor_random_state
        Random seed to use in assigning the remainder neighbors to batches.
        For example, assigning 10 nearest neighbors to 3 batches will make one remainder neighbor randomly assigned to one of the three batches.
        (Default: `0`)
    copy
        Whether to copy the adata or modify in-place.
        (Default: `False`)

    Returns
    ----------
    Union[AnnData, None]
        Depending on `copy`, return an updated or copied :class:`~anndata.AnnData` object with neighborhood graph included.
    """
    #check adata
    adata = adata.copy() if copy else adata
    if batch not in adata.obs:
        raise KeyError(
                f"🛑 '{batch}' is not found in the provided AnnData")
    batch_list = adata.obs[batch].astype(str).values
    if isinstance(cell_type, str) and cell_type not in adata.obs:
        raise KeyError(
                f"🛑 '{cell_type}' is not found in the provided AnnData")
    celltype_list = adata.obs[cell_type].astype(str).values if isinstance(cell_type, str) else np.full(adata.n_obs, 'cell', dtype = object)
    batch_counts = adata.obs[batch].astype(str).value_counts()
    few_batches = set(batch_counts.index[batch_counts <= 10])
    if len(few_batches) > 0:
        logger.warn(f"⚠️ The following batch(es) have too few cells (<= 10), please remove them before running `cellhint.integrate`: {few_batches}")
        return
    if use_rep is None:
        logger.info(f"👀 `use_rep` is not specified, will use `'X_pca'` as the search space")
        use_rep = 'X_pca'
    if use_rep not in adata.obsm.keys():
        raise KeyError(
                f"🛑 '{use_rep}' is not found in `.obsm`")
    n_latent = min([n_latent, adata.obsm[use_rep].shape[1]])
    pca = adata.obsm[use_rep][:, :n_latent]
    #check knn search params
    n_obs = adata.n_obs
    n_neighbors = max([15, 3 * len(np.unique(batch_list))]) if n_neighbors is None else n_neighbors
    swapped = False
    if approx:
        if use_annoy:
            computation = 'annoy'
            if metric not in ['angular', 'euclidean', 'manhattan', 'hamming']:
                swapped = True
                metric = 'euclidean'
        else:
            computation = 'pynndescent'
            if not (metric in pynndescent.distances.named_distances or isinstance(metric, types.FunctionType)):
                swapped = True
                metric = 'euclidean'
    else:
        if not ((metric == 'euclidean') or isinstance(metric, DistanceMetric) or metric in KDTree.valid_metrics):
            swapped = True
            metric = 'euclidean'
        if metric == 'euclidean':
            if 'faiss' in sys.modules and use_faiss:
                computation = 'faiss'
            else:
                computation = 'cKDTree'
        else:
            computation = 'KDTree'
    if swapped:
        logger.warn(f"👀 Unrecognized `metric` for type of neighbor calculation, will switch to 'euclidean'")
    #knn construction
    knn_dists, knn_indices = _get_graph(pca, batch_list, celltype_list, computation, n_neighbors, n_meta_neighbors, metric, annoy_n_trees, pynndescent_n_neighbors, pynndescent_random_state, neighbor_random_state)
    newidx = np.argsort(knn_dists, axis = 1)
    knn_indices = knn_indices[np.arange(n_obs)[:, np.newaxis], newidx]
    knn_dists = knn_dists[np.arange(n_obs)[:, np.newaxis], newidx]
    #connectivities + distances
    X = coo_matrix(([], ([], [])), shape=(n_obs, 1))
    connectivities = fuzzy_simplicial_set(X, n_neighbors, None, None, knn_indices = knn_indices, knn_dists = knn_dists, set_op_mix_ratio = set_op_mix_ratio, local_connectivity = local_connectivity)
    if isinstance(connectivities, tuple):
        connectivities = connectivities[0]
    connectivities = connectivities.tocsr()
    rows = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    cols = np.zeros((n_obs * n_neighbors), dtype=np.int64)
    vals = np.zeros((n_obs * n_neighbors), dtype=np.float64)
    for i in range(n_obs):
        for j in range(n_neighbors):
            if knn_indices[i, j] == -1:
                continue
            if knn_indices[i, j] == i:
                val = 0.0
            else:
                val = knn_dists[i, j]
            rows[i * n_neighbors + j] = i
            cols[i * n_neighbors + j] = knn_indices[i, j]
            vals[i * n_neighbors + j] = val
    distances = coo_matrix((vals, (rows, cols)), shape=(n_obs, n_obs))
    distances.eliminate_zeros()
    #trim
    if trim is None:
        trim = 10 * n_neighbors
    if trim > 0:
        cutoffs = np.zeros(n_obs)
        for i in range(n_obs):
            row_array = connectivities.data[connectivities.indptr[i]: connectivities.indptr[i+1]]
            if row_array.shape[0] <= trim:
                continue
            cutoffs[i] = row_array[np.argsort(row_array)[-1*trim]]
        for iter in range(2):
            for i in range(n_obs):
                row_array = connectivities.data[connectivities.indptr[i]: connectivities.indptr[i+1]]
                row_array[row_array < cutoffs[i]] = 0
            connectivities.eliminate_zeros()
            connectivities = connectivities.T.tocsr()
    #assign
    adata.uns['neighbors'] = {}
    adata.uns['neighbors']['params'] = {'n_neighbors': n_neighbors, 'method': 'umap', 'metric': metric, 'n_pcs': n_latent, 'ccknn': {'n_meta_neighbors': n_meta_neighbors, 'trim': trim, 'computation': computation, 'batch': batch}}
    adata.uns['neighbors']['params']['use_rep'] = use_rep
    adata.obsp['distances'] = distances.tocsr()
    adata.obsp['connectivities'] = connectivities
    adata.uns['neighbors']['distances_key'] = 'distances'
    adata.uns['neighbors']['connectivities_key'] = 'connectivities'
    return adata if copy else None



from anndata import AnnData
from typing import Union, Optional
import pandas as pd
import numpy as np
from .distance import Distance
from .pct import PredictiveClusteringTree
from . import logger

class Distances():
    """
    Class that deals with the cross-dataset cell and cell type distance/dissimilarity analysis.
    For each dataset, a cell-by-cell-type distance matrix is first calculated. This matrix is propagated to other datasets (with the same set of cell types) using predictive clustering trees. Finally, matrices are concatenated across all datasets.

    Parameters
    ----------
    adata
        An :class:`~anndata.AnnData` object containing different datasets/batches and cell types.
        In most scenarios, the format of the expression `.X` in the AnnData is flexible (normalized, log-normalized, z-scaled, etc.).
        However, when `use_rep` is specified as `'X'` (or `X_pca` is not detected in `.obsm` and no other latent representations are provided), `.X` should be log-normalized (to a constant total count per cell).
    dataset
        Column name (key) of cell metadata specifying dataset information.
    cell_type
        Column name (key) of cell metadata specifying cell type information.
    use_rep
        Representation used to calculate distances. This can be `'X'` or any representations stored in `.obsm`.
        Default to the PCA coordinates if present (if not, use the expression matrix `X`).
    metric
        Metric to calculate the distance between each cell and each cell type. Can be `'euclidean'`, `'cosine'`, `'manhattan'` or any metrics applicable to :func:`sklearn.metrics.pairwise_distances`.
        Default to `'euclidean'` if latent representations are used for calculating distances, and to `'correlation'` if the expression matrix is used.
    n_jobs
        Number of CPUs used to calculate intra-dataset distances between cells and cell types. Default to one CPU. `-1` means all CPUs are used.
    **kwargs
        Other keyword arguments passed to :func:`sklearn.metrics.pairwise_distances`.

    Attributes
    ----------
    datasets
        List of datasets involved.
    distances
        List of :class:`~cellhint.distance.Distance` objects, with the ith element for dataset i.
    cell_types
        List of cell types. The ith element lists the cell types in dataset i.
    cells
        List of cells. The ith element lists cells in dataset i.
    adata
        The AnnData that is referred.
    pcts
        List of :class:`~cellhint.pct.PredictiveClusteringTree` objects (after training), with the ith element for dataset i.
    """
    def __init__(self, adata: AnnData, dataset: str, cell_type: str, use_rep: Optional[str] = None, metric: Optional[str] = None, n_jobs: Optional[int] = None, **kwargs):
        if dataset not in adata.obs:
            raise KeyError(
                    f"🛑 '{dataset}' is not found in the provided AnnData")
        if cell_type not in adata.obs:
            raise KeyError(
                    f"🛑 '{cell_type}' is not found in the provided AnnData")
        if use_rep is None:
            if 'X_pca' in adata.obsm.keys():
                logger.info(f"👀 Detected PCA coordinates in the object, will use these to calculate distances")
                use_rep = 'X_pca'
            else:
                logger.info(f"🧙 Using the expression matrix to calculate distances")
                use_rep = 'X'
        elif (use_rep not in adata.obsm.keys()) and (use_rep != 'X'):
            raise KeyError(
                    f"🛑 '{use_rep}' is not found in `.obsm`")
        if use_rep == 'X' and adata.n_vars > 15000:
            logger.warn(f"⚠️ Warning: {adata.n_vars} features are used for calculating distances. Subsetting the AnnData into HVGs is recommended")
        if metric is None:
            metric = 'correlation' if use_rep == 'X' else 'euclidean'
        self.datasets = np.unique(adata.obs[dataset])
        self.distances = []
        for d in self.datasets:
            self.distances.append(Distance.from_adata(adata[adata.obs[dataset] == d], dataset, cell_type, use_rep, metric, n_jobs, False, **kwargs))
        self.use_rep = use_rep
        self.metric = metric
        self.adata = adata

    @property
    def cell_types(self) -> list:
        """Get the cell type list."""
        return [x.cell_type.cell_type.values for x in self.distances]

    @property
    def cells(self) -> list:
        """Get the cell list."""
        return [x.cell.ID.values for x in self.distances]

    def __repr__(self):
        base = f"Within-dataset distance matrices for {len(self.datasets)} datasets"
        base += f"\n    datasets: {str(list(self.datasets))[1:-1]}"
        base += f"\n    distances: list of distances for each dataset"
        if hasattr(self, 'pcts'):
            base += f"\n    pcts: list of predictive clustering trees for each dataset"
        base += f"\n    adata: AnnData object referred"
        return base

    def filter_cells(self, check_symmetry: bool = False) -> None:
        """
        For each dataset, filter out cells whose gene expression profiles do not correlate most with the eigen cell they belong to (i.e., correlate most with other cell types).

        Parameters
        ----------
        check_symmetry
            Whether to check the symmetry of the distance matrix in terms of datasets and cell types.
            (Default: `False`)

        Returns
        ----------
        None
           Modified :class:`~cellhint.distances.Distances` object with undesirable cells filtered out.
        """
        for i in range(len(self.datasets)):
            self.distances[i].filter_cells(check_symmetry)

    def normalize(self, Gaussian_kernel: bool = False, rank: bool = True, normalize: bool = True) -> None:
        """
        Normalize the distance matrix for each dataset with a Gaussian kernel.

        Parameters
        ----------
        Gaussian_kernel
            Whether to apply the Gaussian kernel to the distance matrix.
            (Default: `False`)
        rank
            Whether to turn the matrix into a rank matrx.
            (Default: `True`)
        normalize
            Whether to maximum-normalize the distance matrix.
            (Default: `True`)

        Returns
        ----------
        None
            Modified :class:`~cellhint.distances.Distances` object with normalized intra-dataset distance matrices.
        """
        for i in range(len(self.datasets)):
            self.distances[i].normalize(Gaussian_kernel = Gaussian_kernel, rank = rank, normalize = normalize)

    def train(self,
                max_depth: Optional[int] = None,
                min_samples_split: Union[int, float] = 20,
                min_samples_leaf: Union[int, float] = 10,
                min_weight_fraction_leaf: float = 0.0,
                random_state: Optional[int] = None,
                max_leaf_nodes: Optional[int] = None,
                F_test_prune: bool = True,
                p_thres: float = 0.05,
                sample_weight = None) -> None:
        """
        Train a predictive clustering tree (PCT) for each dataset using the intra-dataset distance matrix.

        Parameters
        ----------
        max_depth
            Maximum possible depth of the tree, starting from the root node which has a depth of 0.
            Default to no limit.
        min_samples_split
            The minimum sample size (in absolute number or fraction) of a possible node.
            (Default: `20`)
        min_samples_leaf
            The minimum sample size (in absolute number or fraction) of a possible leaf.
            (Default: `10`)
        min_weight_fraction_leaf
            The minimum fraction out of total sample weights for a possible leaf.
            (Default: `0.0`)
        random_state
            Random seed for column (feature) shuffling before selecting the best feature and threshold.
        max_leaf_nodes
            The maximum number of leaves, achieved by keeping high-quality (i.e., high impurity reduction) nodes.
            Default to no limit.
        F_test_prune
            Whether to use a F-test to prune the tree by removing unnecessary splits.
            (Default: `True`)
        p_thres
            p-value threshold for pruning nodes after F-test.
            (Default: `0.05`)
        sample_weight
            Sample weights. Default to equal weights across cells.

        Returns
        ----------
        None
            Modified :class:`~cellhint.distances.Distances` object with PCT regressors added as `.pcts`.
        """
        self.pcts = []
        logger.info(f"🏋️ Training the predictive clustering trees for:")
        for da,c,di in zip(self.datasets, self.cells, self.distances):
            logger.info(f"      {da}")
            PCT = PredictiveClusteringTree(max_depth = max_depth,
                                           min_samples_split = min_samples_split,
                                           min_samples_leaf = min_samples_leaf,
                                           min_weight_fraction_leaf = min_weight_fraction_leaf,
                                           random_state = random_state,
                                           max_leaf_nodes = max_leaf_nodes,
                                           F_test_prune = F_test_prune,
                                           p_thres = p_thres)
            PCT.fit(self.adata[c].X, di.dist_mat, sample_weight = sample_weight)
            self.pcts.append(PCT)

    def predict(self, normalize: bool = True, return_distance: bool = True, **kwargs) -> Union[Distance, list]:
        """
        Iteratively predict the cell-by-cell-type distance matrix in other datasets using the predictive clustering tree (PCT) from one dataset.

        Parameters
        ----------
        normalize
            Whether to normalize the distance for each dataset.
            (Default: `True`)
        return_distance
            Whether to join the distances across datasets after prediction to form an entire :class:`~cellhint.distance.Distance` object.
            (Default: `True`)
        **kwargs
            Keyword arguments passed to :func:`~cellhint.distance.Distance.normalize`.

        Returns
        ----------
        If `return_distance` is `True`, a :class:`~cellhint.distance.Distance` object will be returned representing all datasets and cell types combined; otherwise, a list will be returned, with the ith element for dataset i.
        """
        logger.info(f"🖋️ Predicting distances using the PCT in each dataset")
        new_ds = []
        for i in range(len(self.datasets)):
            dist_mat = self.pcts[i].predict(self.adata[np.concatenate(self.cells[:i] + self.cells[i+1:])].X)
            cell = pd.concat([x.cell for x in self.distances[:i]+self.distances[i+1:]], axis = 0)
            pred_dis = Distance(dist_mat, cell, self.distances[i].cell_type)
            new_ds.append(self.distances[i].concatenate(pred_dis, by = 'cell'))
        if normalize:
            for each_new_ds in new_ds:
                each_new_ds.normalize(**kwargs)
        if return_distance:
            return new_ds[0].concatenate(new_ds[1:], by = 'cell_type')
        else:
            return new_ds



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys
sys.path.insert(0, os.path.abspath('../../'))

# -- Required imports --------------------------------------------------------
from datetime import datetime
import cellhint

# -- Project information -----------------------------------------------------
project = 'cellhint'
author = 'Chuan Xu'
copyright = f'{datetime.now():%Y}, {author}'
release = cellhint.__version__

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'myst_parser',
    'nbsphinx'
]

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = ['.rst', '.md']
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'
html_theme_options = {'logo_only': True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_logo = "_static/img/logo_cellhint.png"
html_favicon = "_static/img/logo_cellhint.png"

# -- Extension configuration -------------------------------------------------
napoleon_numpy_docstring = True

# Referring to doctrees in other projects
intersphinx_mapping = dict(
    python=('https://docs.python.org/3/', None),
    numpy=('https://numpy.org/doc/stable/', None),
    pandas=('https://pandas.pydata.org/pandas-docs/stable/', None),
    anndata=('https://anndata.readthedocs.io/en/stable/', None),
    sklearn=('https://scikit-learn.org/stable/', None),
    scanpy=('https://scanpy.readthedocs.io/en/stable/', None),
    plotly=('https://plotly.com/python-api-reference/', None),
    matplotlib=('https://matplotlib.org/stable/', None)
)



from sklearn.metrics import pairwise_distances
import numpy as np
import pandas as pd
from scipy.sparse import spmatrix
from anndata import AnnData
from scipy.stats import rankdata
from typing import Optional
from . import logger
from .symbols import NOVEL, REMAIN, UNASSIGN

class Distance():
    """
    Class that deals with the cross-dataset cell-by-cell-type distance matrix.

    Parameters
    ----------
    dist_mat
        Cell-by-cell-type distance matrix.
    cell
        Cell meta-information including at least `'dataset'`, `'ID'` and `'cell_type'`.
    cell_type
        Cell type meta-information including at least `'dataset'` and `'cell_type'`.

    Attributes
    ----------
    dist_mat
        A cell-by-cell-type distance matrix.
    cell
        Cell meta-information including `'dataset'`, `'ID'` and `'cell_type'`.
    cell_type
        Cell type meta-information including `'dataset'` and `'cell_type'`.
    n_cell
        Number of cells involved.
    n_cell_type
        Number of cell types involved.
    shape
        Tuple of number of cells and cell types.
    assignment
        Assignment of each cell to the most similar cell type in each dataset (obtained through the `assign` method).
    """
    def __init__(self, dist_mat: np.ndarray, cell: pd.DataFrame, cell_type: pd.DataFrame):
        self.dist_mat = dist_mat
        if cell.shape[0] != self.dist_mat.shape[0]:
            raise ValueError(
                    f"🛑 Number of cells in `cell` does not match the cell number in `dist_mat`")
        if cell_type.shape[0] != self.dist_mat.shape[1]:
            raise ValueError(
                    f"🛑 Number of cell types in `cell_type` does not match the cell type number in `dist_mat`")
        if not {'dataset', 'ID', 'cell_type'}.issubset(set(cell.columns)):
            raise KeyError(
                    f"🛑 Please include `'dataset'`, `'ID'` and `'cell_type'` as the cell meta-information")
        if not {'dataset', 'cell_type'}.issubset(set(cell_type.columns)):
            raise KeyError(
                    f"🛑 Please include `'dataset'` and `'cell_type'` as the cell type meta-information")
        self.cell = cell
        self.cell_type = cell_type

    @property
    def n_cell(self) -> int:
        """Number of cells."""
        return self.dist_mat.shape[0]

    @property
    def n_cell_type(self) -> int:
        """Number of cell types."""
        return self.dist_mat.shape[1]

    @property
    def shape(self) -> tuple:
        """Numbers of cells and cell types."""
        return self.dist_mat.shape

    def __repr__(self):
        lend = len(np.unique(self.cell_type.dataset))
        if lend > 1:
            base = f"Cross-dataset distance matrix between {self.n_cell} cells and {self.n_cell_type} cell types from {lend} datasets"
        else:
            base = f"Distance matrix between {self.n_cell} cells and {self.n_cell_type} cell types"
        base += f"\n    dist_mat: distance matrix between {self.n_cell} cells and {self.n_cell_type} cell types"
        base += f"\n    cell: cell meta-information ({str(list(self.cell.columns))[1:-1]})"
        base += f"\n    cell_type: cell type meta-information ({str(list(self.cell_type.columns))[1:-1]})"
        if hasattr(self, 'assignment'):
            base += f"\n    assignment: data frame of cross-dataset cell type assignment"
        return base

    @staticmethod
    def from_adata(adata: AnnData, dataset: str, cell_type: str, use_rep: Optional[str] = None, metric: Optional[str] = None, n_jobs: Optional[int] = None, check_params: bool = True, **kwargs):
        """
        Generate a :class:`~cellhint.distance.Distance` object from the :class:`~anndata.AnnData` given.

        Parameters
        ----------
        adata
            An :class:`~anndata.AnnData` object containing different datasets/batches and cell types.
            In most scenarios, the format of the expression `.X` in the AnnData is flexible (normalized, log-normalized, z-scaled, etc.).
            However, when `use_rep` is specified as `'X'` (or `X_pca` is not detected in `.obsm` and no other latent representations are provided), `.X` should be log-normalized (to a constant total count per cell).
        dataset
            Column name (key) of cell metadata specifying dataset information.
        cell_type
            Column name (key) of cell metadata specifying cell type information.
        use_rep
            Representation used to calculate distances. This can be `'X'` or any representations stored in `.obsm`.
            Default to the PCA coordinates if present (if not, use the expression matrix `X`).
        metric
            Metric to calculate the distance between each cell and each cell type. Can be `'euclidean'`, `'cosine'`, `'manhattan'` or any metrics applicable to :func:`sklearn.metrics.pairwise_distances`.
            Default to `'euclidean'` if latent representations are used for calculating distances, and to `'correlation'` if the expression matrix is used.
        n_jobs
            Number of CPUs used. Default to one CPU. `-1` means all CPUs are used.
        check_params
            Whether to check (or set the default) for `dataset`, `cell_type`, `use_rep` and `metric`.
            (Default: `True`)
        **kwargs
            Other keyword arguments passed to :func:`sklearn.metrics.pairwise_distances`.

        Returns
        ----------
        :class:`~cellhint.distance.Distance`
            A :class:`~cellhint.distance.Distance` object representing the cross-dataset cell-by-cell-type distance matrix.
        """
        #Use `check_params = False` if `dataset`, `cell_type`, `use_rep` and `metric` are already provided correctly.
        if check_params:
            if dataset not in adata.obs:
                raise KeyError(
                        f"🛑 '{dataset}' is not found in the provided AnnData")
            if cell_type not in adata.obs:
                raise KeyError(
                        f"🛑 '{cell_type}' is not found in the provided AnnData")
            if use_rep is None:
                if 'X_pca' in adata.obsm.keys():
                    logger.info(f"👀 Detected PCA coordinates in the object, will use these to calculate distances")
                    use_rep = 'X_pca'
                else:
                    logger.info(f"🧙 Using the expression matrix to calculate distances")
                    use_rep = 'X'
            elif (use_rep not in adata.obsm.keys()) and (use_rep != 'X'):
                raise KeyError(
                        f"🛑 '{use_rep}' is not found in `.obsm`")
            if use_rep == 'X' and adata.n_vars > 15000:
                logger.warn(f"⚠️ Warning: {adata.n_vars} features are used for calculating distances. Subsetting the AnnData into HVGs is recommended")
            if metric is None:
                metric = 'correlation' if use_rep == 'X' else 'euclidean'
        Cell_X = adata.X if use_rep == 'X' else adata.obsm[use_rep]
        IDs = adata.obs_names
        datasets = adata.obs[dataset].astype(str).values
        celltypes = adata.obs[cell_type].astype(str).values
        use_Cell_X = Cell_X if use_rep != 'X' else np.expm1(Cell_X)
        Celltype_X = []
        col_ds = []
        col_cs =[]
        for d in np.unique(datasets):
            for c in np.unique(celltypes[datasets == d]):
                col_cs.append(c)
                col_ds.append(d)
                m = use_Cell_X[(datasets == d) & (celltypes == c), :].mean(axis = 0)
                Celltype_X.append(m.A1 if isinstance(m, np.matrix) else m)
        Celltype_X = np.log1p(np.array(Celltype_X)) if use_rep == 'X' else np.array(Celltype_X)
        if metric not in ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan']:
            if isinstance(Cell_X, spmatrix):
                Cell_X = Cell_X.toarray()
            if isinstance(Celltype_X, spmatrix):
                Celltype_X = Celltype_X.toarray()
        dist_mat = pairwise_distances(Cell_X, Celltype_X, metric = metric, n_jobs = n_jobs, **kwargs)
        cell = pd.DataFrame(dict(dataset=datasets, ID=IDs, cell_type=celltypes))
        cell_type = pd.DataFrame(dict(dataset=col_ds, cell_type=col_cs))
        return Distance(dist_mat, cell, cell_type)

    def normalize(self, Gaussian_kernel: bool = False, rank: bool = True, normalize: bool = True) -> None:
        """
        Normalize the distance matrix with a Gaussian kernel.

        Parameters
        ----------
        Gaussian_kernel
            Whether to apply the Gaussian kernel to the distance matrix.
            (Default: `False`)
        rank
            Whether to turn the matrix into a rank matrx.
            (Default: `True`)
        normalize
            Whether to maximum-normalize the distance matrix.
            (Default: `True`)

        Returns
        ----------
        None
            The :class:`~cellhint.distance.Distance` object modified with a normalized distance matrix.
        """
        if Gaussian_kernel:
            sds = np.sqrt((self.dist_mat ** 2).sum(axis = 1) / self.n_cell_type)[:, np.newaxis]
            self.dist_mat = np.exp(- self.dist_mat / (2 / sds)**2)
            self.dist_mat = 1 - self.dist_mat / self.dist_mat.sum(axis = 1)[:, np.newaxis]
        if rank:
            self.dist_mat = rankdata(self.dist_mat).reshape(self.dist_mat.shape)
        if normalize:
            self.dist_mat = self.dist_mat / self.dist_mat.max()

    def concatenate(self, *distances, by: str = 'cell', check: bool = False):
        """
        Concatenate by either cells (rows) or cell types (columns).

        Parameters
        ----------
        distances
            A :class:`~cellhint.distance.Distance` object or a list of such objects.
        by
            The direction of concatenation, joining either cells (`'cell'`, rows) or cell types (`'cell_type'`, columns).
            (Default: `'cell'`)
        check
            Check whether the concatenation is feasible.
            (Default: `False`)

        Returns
        ----------
        :class:`~cellhint.distance.Distance`
            A :class:`~cellhint.distance.Distance` object concatenated along cells (`by = 'cell'`) or cell types (`by = 'cell_type'`).
        """
        distances = distances[0] if isinstance(distances[0], (list, tuple, set)) else distances
        distances = tuple(distances)
        all_distances = (self,) + distances
        if by not in ['cell', 'cell_type']:
            raise ValueError(
                    f"🛑 Unrecognized `by` value, should be one of `'cell'` or `'cell_type'`")
        if check:
            series_compare = [(x.cell_type.dataset+x.cell_type.cell_type).sort_values() for x in all_distances] if by == 'cell' else [(x.cell.dataset+x.cell.ID).sort_values() for x in all_distances]
            if pd.concat(series_compare, axis = 1).T.drop_duplicates().shape[0] > 1:
                raise Exception(
                        f"🛑 Concatenation is not feasible. Please ensure the meta-information is matched")
        if by == 'cell':
            dist_mat = np.concatenate([x.dist_mat for x in all_distances], axis = 0)
            cell = pd.concat([x.cell for x in all_distances], axis = 0, ignore_index = True)
            return Distance(dist_mat, cell, self.cell_type)
        else:
            match_base = (self.cell.dataset+self.cell.ID).reset_index().set_index(0)
            indices = [np.argsort(match_base.loc[x.cell.dataset+x.cell.ID, 'index'].values) for x in distances]
            dist_mat = np.concatenate([self.dist_mat] + [x.dist_mat[y, :] for x,y in zip(distances, indices)], axis = 1)
            cell_type = pd.concat([x.cell_type for x in all_distances], axis = 0, ignore_index = True)
            return Distance(dist_mat, self.cell, cell_type)

    def symmetric(self) -> bool:
        """
        Check whether the distance matrix is symmetric in terms of datasets and cell types.

        Returns
        ----------
        bool
            `True` or `False` indicating whether all datasets and cell types are included in the object (thus symmetric).
        """
        return np.array_equal(np.unique(self.cell.dataset + self.cell.cell_type), np.unique(self.cell_type.dataset + self.cell_type.cell_type))

    def filter_cells(self, check_symmetry: bool = True) -> None:
        """
        Filter out cells whose gene expression profiles do not correlate most with the eigen cell they belong to (i.e., correlate most with other cell types).

        Parameters
        ----------
        check_symmetry
            Whether to check the symmetry of the distance matrix in terms of datasets and cell types.
            (Default: `True`)

        Returns
        ----------
        None
            A :class:`~cellhint.distance.Distance` object with undesirable cells filtered out.
        """
        if check_symmetry and not self.symmetric():
            raise ValueError(
                    f"🛑 Cell filtering is not possible. Please provide the matrix with symmetric datasets and cell types")
        bool_cell = np.ones(self.n_cell, dtype=bool)
        for i, s in self.cell.iterrows():
            flag_dataset = self.cell_type.dataset == s['dataset']
            if self.cell_type.cell_type.values[flag_dataset][self.dist_mat[i][flag_dataset].argmin()] != s['cell_type']:
                bool_cell[i] = False
        if (~bool_cell).sum() == 0:
            logger.info(f"✂️ No cells are filtered out")
        else:
            ds_unique, ds_table = np.unique(self.cell.dataset.values[~bool_cell], return_counts = True)
            if len(ds_unique) == 1:
                logger.info(f"✂️ {(~bool_cell).sum()} cells are filtered out from {ds_unique[0]}")
            else:
                logger.info(f"✂️ {(~bool_cell).sum()} cells are filtered out, including:")
                for m, n in zip(ds_unique, ds_table):
                    logger.info(f"      {n} cells from {m}")
            self.dist_mat = self.dist_mat[bool_cell]
            self.cell = self.cell[bool_cell]
            all_combine = (self.cell_type.dataset + ': ' + self.cell_type.cell_type).values
            left_combine = np.unique(self.cell.dataset + ': ' + self.cell.cell_type)
            if len(left_combine) < len(all_combine):
                column_keep = np.isin(all_combine, left_combine)
                self.dist_mat = self.dist_mat[:, column_keep]
                self.cell_type = self.cell_type[column_keep]
                logger.info(f"✂️ The following cell types are discarded due to low confidence in annotation:")
                for rec in all_combine[~column_keep]:
                    logger.info(f"      {rec}")

    def to_meta(self, check_symmetry: bool = True, turn_binary: bool = False, return_symmetry: bool = True) -> pd.DataFrame:
        """
        Meta-analysis of cross-dataset cell type dissimilarity or membership.

        Parameters
        ----------
        check_symmetry
            Whether to check the symmetry of the distance matrix in terms of datasets and cell types.
            (Default: `True`)
        turn_binary
            Whether to turn the distance matrix into a cell type membership matrix before meta analysis.
            (Default: `False`)
        return_symmetry
            Whether to return a symmetric dissimilarity matrix by averaging with its transposed form.
            (Default: `True`)

        Returns
        ----------
        :class:`~pandas.DataFrame`
            A :class:`~pandas.DataFrame` object representing the cell-type-level dissimilarity matrix (`turn_binary = False`) or membership matrix (`turn_binary = True`).
        """
        if check_symmetry and not self.symmetric():
            raise ValueError(
                    f"🛑 Meta cell analysis is not possible. Concatenate all datasets and cell types beforehand using `concatenate`")
        use_mat = self.to_binary(False if check_symmetry else True).dist_mat if turn_binary else self.dist_mat
        meta_cell = []
        for _, s in self.cell_type.iterrows():
            meta_cell.append(use_mat[(self.cell.dataset == s['dataset']) & (self.cell.cell_type == s['cell_type']), :].mean(axis = 0))
        meta_cell = pd.DataFrame(np.array(meta_cell))
        meta_cell.index = (self.cell_type.dataset + ': ' + self.cell_type.cell_type).values
        meta_cell.columns = meta_cell.index
        return (meta_cell + meta_cell.T)/2 if return_symmetry else meta_cell

    def to_binary(self, check_symmetry: bool = True):
        """
        Turn the distance matrix into a binary matrix representing the estimated cell type membership across datasets.

        Parameters
        ----------
        check_symmetry
            Whether to check the symmetry of the distance matrix in terms of datasets and cell types.
            (Default: `True`)

        Returns
        ----------
        :class:`~cellhint.distance.Distance`
            A :class:`~cellhint.distance.Distance` object representing the estimated cell type membership across datasets.
        """
        if check_symmetry and not self.symmetric():
            raise ValueError(
                    f"🛑 Cannot convert to a binary matrix. Please provide the matrix with symmetric datasets and cell types")
        member_mat = np.zeros(self.shape, dtype = int)
        datasets = self.cell_type.dataset.values
        for dataset in np.unique(datasets):
            indices = np.where(datasets == dataset)[0]
            member_mat[range(member_mat.shape[0]), indices[self.dist_mat[:, indices].argmin(axis = 1)]] = 1
        return Distance(member_mat, self.cell, self.cell_type)

    def assign(self) -> None:
        """
        Assign each cell to its most similar cell type in each dataset.

        Returns
        ----------
        None
            Modified object with the result of cell assignment added as `.assignment`.
        """
        assignment = {}
        for dataset in np.unique(self.cell_type.dataset):
            flag = self.cell_type.dataset == dataset
            assignment[dataset] = self.cell_type.cell_type.values[flag][self.dist_mat[:, flag].argmin(axis = 1)]
        assignment = pd.DataFrame(assignment, index = self.cell.index)
        #no need to assign cells for the dataset they belong to
        for dataset in assignment.columns:
            flag = self.cell.dataset == dataset
            assignment.loc[flag, dataset] = self.cell.cell_type.values[flag]
        self.assignment = assignment

    def to_confusion(self, D1: str, D2: str, check: bool = True) -> tuple:
        """
        This function is deprecated. Use `to_pairwise_confusion` and `to_multi_confusion` instead.
        Extract the dataset1-by-dataset2 and dataset2-by-dataset1 confusion matrices. Note this function is expected to be applied to a binary membership matrix.

        Parameters
        ----------
        D1
            Name of the first dataset.
        D2
            Name of the second dataset.
        check
            Whether to check names of the two datasets are contained.
            (Default: `True`)

        Returns
        ----------
        tuple
            The dataset1-by-dataset2 and dataset2-by-dataset1 confusion matrices.
        """
        if check and not {D1, D2}.issubset(np.unique(self.cell_type.dataset)):
            raise ValueError(
                    f"🛑 Please provide correct dataset names")
        D1_col_flag = self.cell_type.dataset == D1
        D2_col_flag = self.cell_type.dataset == D2
        D1_celltypes = self.cell_type.cell_type.values[D1_col_flag]
        D2_celltypes = self.cell_type.cell_type.values[D2_col_flag]
        D1_row_flag = self.cell.dataset == D1
        D2_row_flag = self.cell.dataset == D2
        D1byD2 = pd.DataFrame(np.array([self.dist_mat[D1_row_flag & (self.cell.cell_type == x)][:, D2_col_flag].sum(axis=0) for x in D1_celltypes]), columns = D2_celltypes, index = D1_celltypes)
        D2byD1 = pd.DataFrame(np.array([self.dist_mat[D2_row_flag & (self.cell.cell_type == x)][:, D1_col_flag].sum(axis=0) for x in D2_celltypes]), columns = D1_celltypes, index = D2_celltypes)
        return D1byD2, D2byD1

    def to_pairwise_confusion(self, D1: str, D2: str, check: bool = True) -> tuple:
        """
        Extract the dataset1-by-dataset2 and dataset2-by-dataset1 confusion matrices.

        Parameters
        ----------
        D1
            Name of the first dataset.
        D2
            Name of the second dataset.
        check
            Whether to check names of the two datasets are contained.
            (Default: `True`)

        Returns
        ----------
        tuple
            The dataset1-by-dataset2 and dataset2-by-dataset1 confusion matrices.
        """
        if check and not {D1, D2}.issubset(np.unique(self.cell_type.dataset)):
            raise ValueError(
                    f"🛑 Please provide correct dataset names")
        if not hasattr(self, 'assignment'):
            raise AttributeError(
                    f"🛑 No `.assignment` attribute in the object. Use the `.assign` method first")
        D1_flag = (self.cell.dataset == D1)
        D2_flag = (self.cell.dataset == D2)
        D1byD2 = pd.crosstab(self.cell.cell_type[D1_flag], self.assignment.loc[D1_flag, D2])
        D2byD1 = pd.crosstab(self.cell.cell_type[D2_flag], self.assignment.loc[D2_flag, D1])
        D1byD2_lack_columns = D2byD1.index.difference(D1byD2.columns)
        if len(D1byD2_lack_columns) > 0:
            D1byD2 = D1byD2.join(pd.DataFrame(np.zeros((len(D1byD2.index), len(D1byD2_lack_columns)), dtype=int), index = D1byD2.index, columns = D1byD2_lack_columns))
        D2byD1_lack_columns = D1byD2.index.difference(D2byD1.columns)
        if len(D2byD1_lack_columns) > 0:
            D2byD1 = D2byD1.join(pd.DataFrame(np.zeros((len(D2byD1.index), len(D2byD1_lack_columns)), dtype=int), index = D2byD1.index, columns = D2byD1_lack_columns))
        return D1byD2, D2byD1.loc[D1byD2.columns, D1byD2.index]

    def to_multi_confusion(self, relation: pd.DataFrame, D: str, check: bool = True) -> tuple:
        """
        Extract the confusion matrices between meta-cell-types defined prior and cell types from a new dataset.

        Parameters
        ----------
        relation
            A :class:`~pandas.DataFrame` object representing the cell type harmonization result across multiple datasets.
        D
            Name of the new dataset to be aligned.
        check
            Whether to check names of the datasets are contained.
            (Default: `True`)

        Returns
        ----------
        tuple
            The confusion matrices between meta-cell-types defined prior and cell types from a new dataset.
        """
        datasets = relation.columns[0::2]
        if check:
            if not set(datasets).issubset(np.unique(self.cell_type.dataset)):
                raise ValueError(
                        f"🛑 `relation` contains unexpected dataset names")
            if D not in np.unique(self.cell_type.dataset) or D in datasets:
                raise ValueError(
                        f"🛑 Please provide a valid dataset name `D`")
        if not hasattr(self, 'assignment'):
            raise AttributeError(
                    f"🛑 No `.assignment` attribute in the object. Use the `.assign` method first")
        #D1byD2
        D1_flag = self.cell.dataset.isin(datasets)
        D1_assign = self.assignment[D1_flag]
        D1_truth = np.full(D1_assign.shape[0], UNASSIGN, dtype = object)
        for _, s in relation.iterrows():
            celltypes = s.values[0::2]
            non_blank_flag = ~np.isin(celltypes, [NOVEL, REMAIN])
            existing_datasets = datasets[non_blank_flag]
            existing_celltypes = celltypes[non_blank_flag]
            flag = np.all(D1_assign[existing_datasets] == existing_celltypes, axis = 1).values & self.cell[D1_flag].dataset.isin(existing_datasets).values
            D1_truth[flag] = ' '.join(s.values)
        D1_used = D1_truth != UNASSIGN
        D1byD2 = pd.crosstab(D1_truth[D1_used], D1_assign.loc[D1_used, D])
        #D2byD1
        D2_flag = self.cell.dataset == D
        D2_assign = self.assignment[D2_flag]
        D2_predict = np.full(D2_assign.shape[0], UNASSIGN, dtype = object)
        for _, s in relation.iterrows():
            celltypes = s.values[0::2]
            flags = (D2_assign[datasets] == celltypes) | np.isin(celltypes, [NOVEL, REMAIN])
            D2_predict[np.all(flags, axis = 1).values] = ' '.join(s.values)
        D2_used = D2_predict != UNASSIGN
        D2byD1 = pd.crosstab(self.cell.cell_type[D2_flag][D2_used], D2_predict[D2_used])
        #warning
        if relation.shape[0] > D1byD2.shape[0]:
            lost_celltypes = np.setdiff1d(relation.apply(lambda row: ' '.join(row.values), axis = 1).values, D1byD2.index)
            logger.warn(f"⚠️ Warning: no cells are found to match these patterns: {set(lost_celltypes)}. Double check the harmonized relationships before integrating '{D}'")
            D1byD2 = pd.concat([D1byD2, pd.DataFrame(np.zeros((len(lost_celltypes), len(D1byD2.columns)), dtype=int), index = lost_celltypes, columns = D1byD2.columns)], axis = 0)
        #a unique cell type in D2 may be annotated to nothing and filtered
        lost_celltypes = np.setdiff1d(np.unique(self.cell.cell_type[D2_flag]), D2byD1.index)
        if len(lost_celltypes) > 0:
            D2byD1 = pd.concat([D2byD1, pd.DataFrame(np.zeros((len(lost_celltypes), len(D2byD1.columns)), dtype=int), index = lost_celltypes, columns = D2byD1.columns)], axis = 0)
        #return
        D1byD2_lack_columns = D2byD1.index.difference(D1byD2.columns)
        if len(D1byD2_lack_columns) > 0:
            D1byD2 = D1byD2.join(pd.DataFrame(np.zeros((len(D1byD2.index), len(D1byD2_lack_columns)), dtype=int), index = D1byD2.index, columns = D1byD2_lack_columns))
        D2byD1_lack_columns = D1byD2.index.difference(D2byD1.columns)
        if len(D2byD1_lack_columns) > 0:
            D2byD1 = D2byD1.join(pd.DataFrame(np.zeros((len(D2byD1.index), len(D2byD1_lack_columns)), dtype=int), index = D2byD1.index, columns = D2byD1_lack_columns))
        return D1byD2, D2byD1.loc[D1byD2.columns, D1byD2.index]



import numpy as np
import pandas as pd
from typing import Union, Optional
from . import logger
from .distance import Distance
from .symbols import ONE2ONE, ONE2MANY, MANY2ONE, NOVEL, REMAIN, UNASSIGN
from .plot import SEP1, _identify_relation_groups
import os
import pickle

def _subset_and_normalize(D1byD2: pd.DataFrame, D2byD1: pd.DataFrame, relation: pd.DataFrame, row_normalize: bool = True) -> tuple:
    """
    For internal use. Subset the two matrices and row normalize to a sum of 1.
    """
    if relation.shape[0] > 0:
        flag1 = ~D1byD2.index.isin(relation.D1.unique())
        flag2 = ~D1byD2.columns.isin(relation.D2.unique())
        D1byD2 = D1byD2.loc[flag1, flag2]
        D2byD1 = D2byD1.loc[flag2, flag1]
    if not D1byD2.empty and row_normalize:
        D1byD2 = D1byD2 / D1byD2.sum(axis = 1).values[:, np.newaxis]
        D2byD1 = D2byD1 / D2byD1.sum(axis = 1).values[:, np.newaxis]
    return D1byD2, D2byD1

def _pairwise_align(D1byD2: pd.DataFrame, D2byD1: pd.DataFrame, check: bool = True, row_normalize: bool = True, minimum_unique_percent: float = 0.5, minimum_divide_percent: float = 0.1, maximum_novel_percent: float = 0.05) -> pd.DataFrame:
    """
    For internal use. Align cell types from two datasets based on two confusion matrices.
    """
    if check and (not np.array_equal(D1byD2.index, D2byD1.columns) or not np.array_equal(D1byD2.columns, D2byD1.index)):
        raise ValueError(
                f"🛑 Matrices are not provided correctly")
    relation = pd.DataFrame(columns = ['D1', 'relation', 'D2'])
    #rows all 0
    rownames = D1byD2.index[D1byD2.sum(axis = 1) == 0]
    if len(rownames) > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': rownames, 'relation': [ONE2ONE]*len(rownames), 'D2': [NOVEL]*len(rownames)})])
    rownames = D2byD1.index[D2byD1.sum(axis = 1) == 0]
    if len(rownames) > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': [NOVEL]*len(rownames), 'relation': [ONE2ONE]*len(rownames), 'D2': rownames})])
    #subset and normalize
    D1byD2, D2byD1 = _subset_and_normalize(D1byD2, D2byD1, relation, row_normalize)
    #D1 novel
    colnames = D2byD1.columns[D2byD1.max(axis = 0) < maximum_novel_percent]
    if len(colnames) > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': colnames, 'relation': [ONE2ONE]*len(colnames), 'D2': [NOVEL]*len(colnames)})])
    #D2 novel
    colnames = D1byD2.columns[D1byD2.max(axis = 0) < maximum_novel_percent]
    if len(colnames) > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': [NOVEL]*len(colnames), 'relation': [ONE2ONE]*len(colnames), 'D2': colnames})])
    D1byD2, D2byD1 = _subset_and_normalize(D1byD2, D2byD1, relation, row_normalize)
    #one2one
    indices = (D1byD2 > minimum_unique_percent) & (D2byD1.T > minimum_unique_percent)
    one2many_indices = (D1byD2 > minimum_divide_percent) & (D2byD1.T > minimum_unique_percent)
    one2many_indices = one2many_indices & (one2many_indices.sum(axis = 1) >= 2).values[:, np.newaxis]
    many2one_indices = (D2byD1 > minimum_divide_percent) & (D1byD2.T > minimum_unique_percent)
    many2one_indices = many2one_indices & (many2one_indices.sum(axis = 1) >= 2).values[:, np.newaxis]
    row_index, col_index = np.where(indices & (indices.sum(axis = 1) == 1).values[:, np.newaxis] & (indices.sum(axis = 0) == 1).values & (~one2many_indices) & (~many2one_indices.T))
    while len(row_index) > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': D1byD2.index[row_index], 'relation': [ONE2ONE]*len(row_index), 'D2': D1byD2.columns[col_index]})])
        D1byD2, D2byD1 = _subset_and_normalize(D1byD2, D2byD1, relation, row_normalize)
        if D1byD2.empty:
            break
        indices = (D1byD2 > minimum_unique_percent) & (D2byD1.T > minimum_unique_percent)
        one2many_indices = (D1byD2 > minimum_divide_percent) & (D2byD1.T > minimum_unique_percent)
        one2many_indices = one2many_indices & (one2many_indices.sum(axis = 1) >= 2).values[:, np.newaxis]
        many2one_indices = (D2byD1 > minimum_divide_percent) & (D1byD2.T > minimum_unique_percent)
        many2one_indices = many2one_indices & (many2one_indices.sum(axis = 1) >= 2).values[:, np.newaxis]
        row_index, col_index = np.where(indices & (indices.sum(axis = 1) == 1).values[:, np.newaxis] & (indices.sum(axis = 0) == 1).values & (~one2many_indices) & (~many2one_indices.T))
    #one2many
    if not D1byD2.empty:
        indices = (D1byD2 > minimum_divide_percent) & (D2byD1.T > minimum_unique_percent)
        row_index, col_index = np.where(indices & (indices.sum(axis = 1) >= 2).values[:, np.newaxis])
        while len(row_index) > 0:
            relation = pd.concat([relation, pd.DataFrame({'D1': D1byD2.index[row_index], 'relation': [ONE2MANY]*len(row_index), 'D2': D1byD2.columns[col_index]})])
            D1byD2, D2byD1 = _subset_and_normalize(D1byD2, D2byD1, relation, row_normalize)
            if D1byD2.empty:
                break
            indices = (D1byD2 > minimum_divide_percent) & (D2byD1.T > minimum_unique_percent)
            row_index, col_index = np.where(indices & (indices.sum(axis = 1) >= 2).values[:, np.newaxis])
    #many2one
    if not D1byD2.empty:
        indices = (D2byD1 > minimum_divide_percent) & (D1byD2.T > minimum_unique_percent)
        row_index, col_index = np.where(indices & (indices.sum(axis = 1) >= 2).values[:, np.newaxis])
        while len(row_index) > 0:
            relation = pd.concat([relation, pd.DataFrame({'D1': D2byD1.columns[col_index], 'relation': [MANY2ONE]*len(col_index), 'D2': D2byD1.index[row_index]})])
            D1byD2, D2byD1 = _subset_and_normalize(D1byD2, D2byD1, relation, row_normalize)
            if D1byD2.empty:
                break
            indices = (D2byD1 > minimum_divide_percent) & (D1byD2.T > minimum_unique_percent)
            row_index, col_index = np.where(indices & (indices.sum(axis = 1) >= 2).values[:, np.newaxis])
    #remaining
    if D1byD2.shape[0] > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': D1byD2.index, 'relation': [ONE2ONE]*len(D1byD2.index), 'D2': [REMAIN]*len(D1byD2.index)})])
    if D1byD2.shape[1] > 0:
        relation = pd.concat([relation, pd.DataFrame({'D1': [REMAIN]*len(D1byD2.columns), 'relation': [ONE2ONE]*len(D1byD2.columns), 'D2': D1byD2.columns})])
    return relation

def _reorder_dataset(sim_df: pd.DataFrame) -> np.ndarray:
    """
    For internal use. Reorder datasets based on their pairwise similarities.
    """
    stack_datasets = sim_df.iloc[sim_df.similarity.argmax(), [0, 1]].tolist()
    sim_df = pd.concat([sim_df, sim_df[['D2', 'D1', 'similarity']].rename(columns = {'D2':'D1', 'D1':'D2'})])
    remain_datasets = list(np.setdiff1d(np.unique(sim_df.D1), stack_datasets))
    while len(remain_datasets) >= 2:
        scores = [sim_df.loc[(sim_df.D1 == x) & sim_df.D2.isin(stack_datasets), 'similarity'].sum() for x in remain_datasets]
        stack_datasets.append(remain_datasets.pop(scores.index(max(scores))))
    stack_datasets.append(remain_datasets.pop())
    return np.array(stack_datasets, dtype = object)

class DistanceAlignment():
    """
    Class that performs cell type label harmonization across datasets.

    Parameters
    ----------
    base_distance
        A :class:`~cellhint.distance.Distance` object.
    check
        Whether to check the supplied `base_distance` is correctly provided.
        (Default: `True`)
    dataset_order
        Order of datasets to be aligned. By default, the order is the same as that in the base distance matrix.
    row_normalize
        Whether to row normalize the confusion matrix to a sum of 1 in each iteration.
        (Default: `True`)
    minimum_unique_percent
        The minimum cell assignment fraction to claim a cell type as uniquely matched to a cell type from the other dataset.
        (Default: `0.5`)
    minimum_divide_percent
        The minimum cell assignment fraction to claim a cell type as divisible into two or more cell types from the other dataset.
        (Default: `0.1`)
    maximum_novel_percent
        The maximum cell assignment fraction to claim a cell type as novel to a given dataset.
        (Default: `0.05`)

    Attributes
    ----------
    base_distance
        The :class:`~cellhint.distance.Distance` object.
    dataset_order
        Order of datasets to be aligned.
    row_normalize
        Whether to row normalize the confusion matrix to a sum of 1 in each iteration.
    minimum_unique_percent
        The minimum cell assignment fraction to claim a cell type as uniquely matched to a cell type from the other dataset.
    minimum_divide_percent
        The minimum cell assignment fraction to claim a cell type as divisible into two or more cell types from the other dataset.
    maximum_novel_percent
        The maximum cell assignment fraction to claim a cell type as novel to a given dataset.
    relation
        A :class:`~pandas.DataFrame` representing the harmonization result.
    aligned_datasets
        List of datasets that are already harmonized.
    groups
        Cell type groups (high-hierarchy cell types) categorizing the rows of `.relation`.
    reannotation
        A :class:`~pandas.DataFrame` representing the reannotated cell types.
    minimum_unique_percents
        List of `minimum_unique_percent` values which are used along harmonization iterations in order to get the best alignment.
        This attribute is obtained through the :func:`~cellhint.align.DistanceAlignment.best_align` function.
    minimum_divide_percents
        List of `minimum_divide_percent` values which are used along harmonization iterations in order to get the best alignment.
        This attribute is obtained through the :func:`~cellhint.align.DistanceAlignment.best_align` function.
    """
    def __init__(self, base_distance: Distance, check: bool = True, dataset_order: Optional[Union[list, tuple, np.ndarray, pd.Series, pd.Index]] = None,
                 row_normalize: bool = True, minimum_unique_percent: float = 0.5, minimum_divide_percent: float = 0.1, maximum_novel_percent: float = 0.05):
        if check:
            if not isinstance(base_distance, Distance) or not base_distance.symmetric():
                raise ValueError(
                        f"🛑 Please provide a symmetric `Distance` object")
            if not hasattr(base_distance, 'assignment'):
                raise AttributeError(
                        f"🛑 No `.assignment` attribute in the `base_distance`. Apply the `.assign` method first")
        self.base_distance = base_distance
        if dataset_order is None:
            dataset_order = np.unique(self.base_distance.cell_type.dataset)
        else:
            dataset_order = np.array(dataset_order, dtype = object)
            if not np.array_equal(np.sort(dataset_order), np.unique(self.base_distance.cell_type.dataset)):
                raise ValueError(
                        f"🛑 Please provide a comprehensive order of datasets with correct names")
        self.dataset_order = dataset_order
        self.row_normalize = row_normalize
        self.minimum_unique_percent = minimum_unique_percent
        self.minimum_divide_percent = minimum_divide_percent
        self.maximum_novel_percent = maximum_novel_percent

    def reorder_dataset(self, weights: Union[list, tuple, np.ndarray, pd.Series, pd.Index] = (2, 1, -1, -2), return_similarity: bool = False) -> Union[None, pd.DataFrame]:
        """
        Reorder the datasets such that similar datasets will be harmonized first. This method can also be used to calculate CellHint-defined inter-dataset similarities.

        Parameters
        ----------
        weights
            Weights assigned to one-to-one, one/many-to-many/one, novel, and remaining cell type matches, respectively. Default to 2, 1, -1, -2.
            Inter-cell-type similarities will be weighted by these values to derive the weighted sum of similarity between each pair of datasets.
        return_similarity
            Whether to return the data frame of dataset-dataset similarities.
            (Default: `False`)

        Returns
        ----------
        Reordered datasets as the attribute `.dataset_order` and if `return_similarity = True`, return a :class:`~pandas.DataFrame` of dataset-dataset similarities.
        """
        if len(self.dataset_order) == 2:
            logger.warn(f"⚠️ Warning: only two datasets exist, no need to reorder them")
            return
        weights = np.array(weights, dtype = 'float')
        meta_distance = self.base_distance.to_meta(False)
        meta_similarity = 1 - meta_distance
        sim_df = pd.DataFrame([self.dataset_order[[i, j]] for i in range(0, len(self.dataset_order) - 1) for j in range(i + 1, len(self.dataset_order))], columns = ['D1', 'D2'])
        scores = []
        for _, s in sim_df.iterrows():
            D1 = s.values[0]
            D2 = s.values[1]
            sub_meta_similarity = meta_similarity.loc[meta_similarity.index.str.startswith(D1 + ": "), meta_similarity.columns.str.startswith(D2 + ": ")]
            sub_meta_similarity.index = sub_meta_similarity.index.str.replace(D1 + ": ", '', regex = False)
            sub_meta_similarity.columns = sub_meta_similarity.columns.str.replace(D2 + ": ", '', regex = False)
            relation = self.pairwise_align(D1, D2, False)
            ss = np.full(relation.shape[0], -1, dtype = 'float')
            ws = np.full(relation.shape[0], weights[0], dtype = 'float')
            #REMAIN NOVEL
            flag_REMAIN_1 = (relation[D1] == REMAIN).values
            flag_REMAIN_2 = (relation[D2] == REMAIN).values
            flag_NOVEL_1 = (relation[D1] == NOVEL).values
            flag_NOVEL_2 = (relation[D2] == NOVEL).values
            flag_NONE_1 = flag_REMAIN_1 | flag_NOVEL_1
            if flag_NONE_1.sum() > 0:
                ss[flag_NONE_1] = 1 - sub_meta_similarity[relation.loc[flag_NONE_1, D2].values].max(axis = 0).values
            flag_NONE_2 = flag_REMAIN_2 | flag_NOVEL_2
            if flag_NONE_2.sum() > 0:
                ss[flag_NONE_2] = 1 - sub_meta_similarity.loc[relation.loc[flag_NONE_2, D1].values].max(axis = 1).values
            ws[flag_REMAIN_1 | flag_REMAIN_2] = weights[3]
            ws[flag_NOVEL_1 | flag_NOVEL_2] = weights[2]
            #other ss
            other_flag = ~(flag_NONE_1 | flag_NONE_2)
            if other_flag.sum() > 0:
                ss[other_flag] = np.diag(sub_meta_similarity.loc[relation.loc[other_flag, D1].values, relation.loc[other_flag, D2].values])
            ##remove in the future-->
            assert np.all(ss != -1)
            ##<<-remove in the future
            #ONE2MANY and MANY2ONE weights
            flag_ONE2MANY = (relation.relation == ONE2MANY).values
            if flag_ONE2MANY.sum() > 0:
                DS = relation.loc[flag_ONE2MANY, D1].value_counts()
                ws[flag_ONE2MANY] = (weights[1] / DS)[relation.loc[flag_ONE2MANY, D1].values].values
            flag_MANY2ONE = (relation.relation == MANY2ONE).values
            if flag_MANY2ONE.sum() > 0:
                DS = relation.loc[flag_MANY2ONE, D2].value_counts()
                ws[flag_MANY2ONE] = (weights[1] / DS)[relation.loc[flag_MANY2ONE, D2].values].values
            #final scores
            scores.append(np.sum(ss * ws) / sum(sub_meta_similarity.shape))
        sim_df['similarity'] = scores
        self.dataset_order = _reorder_dataset(sim_df)
        if return_similarity:
            return sim_df

    def pairwise_align(self, D1: str, D2: str, check: bool = True) -> pd.DataFrame:
        """
        Pairwise alignment of cell types between two datasets.

        Parameters
        ----------
        D1
            Name of the first dataset.
        D2
            Name of the second dataset.
        check
            Whether to check names of the two datasets are contained in the :attr:`~cellhint.align.DistanceAlignment.base_distance`.
            (Default: `True`)

        Returns
        ----------
        :class:`~pandas.DataFrame`
            A :class:`~pandas.DataFrame` with three columns:
            1) **name of dataset 1**, cell types from dataset 1.
            2) **relation**, being either '=', '∋' or '∈'.
            3) **name of dataset 2**, cell types from dataset 2.
        """
        D1byD2, D2byD1 = self.base_distance.to_pairwise_confusion(D1, D2, check)
        relation = _pairwise_align(D1byD2, D2byD1, False, self.row_normalize, self.minimum_unique_percent, self.minimum_divide_percent, self.maximum_novel_percent)
        return relation.rename(columns = {'D1': D1, 'D2': D2})

    def multi_align(self, relation: pd.DataFrame, D: str, check: bool = True) -> pd.DataFrame:
        """
        Multiple alignment of cell types across datasets. Cell types from a new dataset will be integrated into the previous harmonization data frame.

        Parameters
        ----------
        relation
            A :class:`~pandas.DataFrame` object representing the cell type harmonization result across multiple datasets.
        D
            Name of the new dataset to be aligned.
        check
            Whether to check names of the datasets are contained.
            (Default: `True`)

        Returns
        ----------
        :class:`~pandas.DataFrame`
            A :class:`~pandas.DataFrame` with multiple columns:
            1) **name of dataset 1**, cell types from dataset 1.
            2) **relation**, being either '=', '∋' or '∈'.
            3) **name of dataset 2**, cell types from dataset 2.
            4) ...
            N) **name of the new dataset**, cell types from the new dataset.
        """
        D1byD2, D2byD1 = self.base_distance.to_multi_confusion(relation, D, check)
        new_relation = _pairwise_align(D1byD2, D2byD1, False, self.row_normalize, self.minimum_unique_percent, self.minimum_divide_percent, self.maximum_novel_percent)
        relation.index = relation.apply(lambda row: ' '.join(row.values), axis = 1).values
        relation.loc[NOVEL] = np.tile([NOVEL, ONE2ONE], int(relation.shape[1]/2)+1)[:-1]
        relation.loc[REMAIN] = np.tile([REMAIN, ONE2ONE], int(relation.shape[1]/2)+1)[:-1]
        relation = relation.loc[new_relation.D1.values]
        relation.index = new_relation.index
        relation['__relation'] = new_relation['relation'].values
        relation[D] = new_relation['D2'].values
        return relation.rename(columns = {'__relation': 'relation'})

    def align(self, datasets: Optional[Union[list, tuple, np.ndarray, pd.Series, pd.Index]] = None) -> None:
        """
        Iterative alignment of cell types across datasets.

        Parameters
        ----------
        datasets
            Datasets to be aligned. Default to using all datasets available.

        Returns
        ----------
        None
            A :class:`~pandas.DataFrame` with multiple columns added as the attribute `.relation`:
            1) **name of dataset 1**, cell types from dataset 1.
            2) **relation**, being either '=', '∋' or '∈'.
            3) **name of dataset 2**, cell types from dataset 2.
            4) ...
            N) **name of the last dataset**, cell types from the last dataset.
        """
        if datasets is None:
            datasets = self.dataset_order
        else:
            datasets = np.array(datasets, dtype = object)
            if not set(datasets).issubset(self.dataset_order):
                raise ValueError(
                        f"🛑 The following datasets are not found: {set(datasets).difference(self.dataset_order)}")
        logger.info(f"🖇 Harmonizing cell types of {datasets[0]} and {datasets[1]}")
        relation = self.pairwise_align(datasets[0], datasets[1], False)
        if len(datasets) >= 3:
            for i in range(2, len(datasets)):
                logger.info(f"🖇 Harmonizing cell types of {datasets[i]}")
                relation = self.multi_align(relation, datasets[i], False)
        logger.info(f"✅ Harmonization done!")
        self.relation = relation

    @property
    def aligned_datasets(self) -> np.ndarray:
        """Get the datasets which are already harmonized."""
        return np.array(self.relation.columns[0::2])

    @property
    def groups(self) -> np.ndarray:
        """Get the cell type groups (high hierarchy) based on the relation table."""
        return _identify_relation_groups(self.relation, order_row = False, order_column = False)[0]

    def __repr__(self):
        base = f"Cross-dataset cell type alignment for {len(self.dataset_order)} datasets"
        base += f"\n    base_distance: a cross-dataset distance object"
        if hasattr(self, 'relation'):
            base += f"\n    aligned_datasets: {str(list(self.aligned_datasets))[1:-1]}"
            base += f"\n    relation: data frame of the harmonization table"
        if hasattr(self, 'reannotation'):
            base += f"\n    reannotation: data frame of the reannotated cells ({str(list(self.reannotation.columns))[1:-1]})"
        return base

    def update(self, datasets: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]] = None) -> None:
        """
        Iteratively update the alignment of cell types across datasets.

        Parameters
        ----------
        datasets
            Datasets to be aligned. Default to using all the remaining datasets.

        Returns
        ----------
        None
            An updated :class:`~pandas.DataFrame` with multiple columns added as the attribute `.relation`:
            1) **name of dataset 1**, cell types from dataset 1.
            2) **relation**, being either '=', '∋' or '∈'.
            3) **name of dataset 2**, cell types from dataset 2.
            4) ...
            N) **name of the last dataset**, cell types from the last dataset.
        """
        if not hasattr(self, 'relation'):
            raise AttributeError(
                    f"🛑 No harmonization result exists. Please run the `.align` method first")
        if len(self.aligned_datasets) == len(self.dataset_order):
            logger.warn(f"⚠️ All datasets have been harmonized. No update is needed")
            return
        remaining_datasets = self.dataset_order[~np.isin(self.dataset_order, self.aligned_datasets)]
        if datasets is None:
            datasets = remaining_datasets
        else:
            datasets = [datasets] if isinstance(datasets, str) else np.array(datasets, dtype = object)
            if not set(datasets).issubset(remaining_datasets):
                raise ValueError(
                        f"🛑 Please provide dataset names from the following list: {set(remaining_datasets)}")
        for dataset in datasets:
            logger.info(f"🖇 Harmonizing cell types of {dataset}")
            self.relation = self.multi_align(self.relation, dataset, False)
        logger.info(f"✅ Harmonization done!")

    def best_align(self, dataset_order: Optional[Union[list, tuple, np.ndarray, pd.Series, pd.Index]] = None, minimum_unique_percents: Union[list, tuple, np.ndarray, pd.Series, pd.Index, float] = (0.4, 0.5, 0.6, 0.7, 0.8), minimum_divide_percents: Union[list, tuple, np.ndarray, pd.Series, pd.Index, float] = (0.1, 0.15, 0.2)):
        """
        Iterative alignment of cell types across datasets by finding the best parameter combo in each iteration.

        Parameters
        ----------
        dataset_order
            Order of datasets to be aligned. This can also be a subset of datasets.
            Default to the dataset order in the `DistanceAlignment` object.
        minimum_unique_percents
            The minimum cell assignment fraction(s) to claim a cell type as uniquely matched to a cell type from the other dataset.
            By default, five values will be tried (0.4, 0.5, 0.6, 0.7, 0.8) to find the one that produces least alignments in each harmonization iteration.
        minimum_divide_percents
            The minimum cell assignment fraction(s) to claim a cell type as divisible into two or more cell types from the other dataset.
            By default, three values will be tried (0.1, 0.15, 0.2) to find the one that produces least alignments in each harmonization iteration.

        Returns
        ----------
        None
            A :class:`~pandas.DataFrame` with multiple columns added as the attribute `.relation`:
            1) **name of dataset 1**, cell types from dataset 1.
            2) **relation**, being either '=', '∋' or '∈'.
            3) **name of dataset 2**, cell types from dataset 2.
            4) ...
            N) **name of the last dataset**, cell types from the last dataset.
        """
        if dataset_order is not None:
            dataset_order = np.array(dataset_order, dtype = object)
            if not set(dataset_order).issubset(self.dataset_order):
                raise ValueError(
                        f"🛑 The following dataset(s) are not found: {set(dataset_order).difference(self.dataset_order)}")
        else:
            dataset_order = self.dataset_order
        original_mup = self.minimum_unique_percent
        original_mdp = self.minimum_divide_percent
        minimum_unique_percents = np.array([minimum_unique_percents]) if isinstance(minimum_unique_percents, float) else np.array(minimum_unique_percents)
        minimum_divide_percents = np.array([minimum_divide_percents]) if isinstance(minimum_divide_percents, float) else np.array(minimum_divide_percents)
        mups = np.full(len(dataset_order) - 1, -1, dtype = 'float')
        mdps = mups.copy()
        logger.info(f"🖇 Harmonizing cell types of {dataset_order[0]} and {dataset_order[1]}")
        n_rows = np.inf
        for minimum_unique_percent in minimum_unique_percents:
            for minimum_divide_percent in minimum_divide_percents:
                self.minimum_unique_percent = minimum_unique_percent
                self.minimum_divide_percent = minimum_divide_percent
                per_relation = self.pairwise_align(dataset_order[0], dataset_order[1], check = False)
                if per_relation.shape[0] <= n_rows:
                    relation = per_relation
                    n_rows = per_relation.shape[0]
                    mups[0] = minimum_unique_percent
                    mdps[0] = minimum_divide_percent
        if len(dataset_order) >= 3:
            for i in range(2, len(dataset_order)):
                logger.info(f"🖇 Harmonizing cell types of {dataset_order[i]}")
                n_rows = np.inf
                for minimum_unique_percent in minimum_unique_percents:
                    for minimum_divide_percent in minimum_divide_percents:
                        self.minimum_unique_percent = minimum_unique_percent
                        self.minimum_divide_percent = minimum_divide_percent
                        per_relation = self.multi_align(relation.copy(), dataset_order[i], check = False)
                        if per_relation.shape[0] <= n_rows:
                            expand_relation = per_relation
                            n_rows = per_relation.shape[0]
                            mups[i - 1] = minimum_unique_percent
                            mdps[i - 1] = minimum_divide_percent
                relation = expand_relation
        self.minimum_unique_percents = mups
        self.minimum_divide_percents = mdps
        self.minimum_unique_percent = original_mup
        self.minimum_divide_percent = original_mdp
        self.relation = relation

    #def reannotate(self, show_iteration: bool = False, add_group: bool = True, prefix: str = '') -> None:
    #    """
    #    Reannotate each cell into the harmonized cell type.

    #    Parameters
    #    ----------
    #    show_iteration
    #        Whether to store the cell type reannotation result for each harmonization iteration.
    #        (Default: `False`)
    #    add_group
    #        Whether to annotate out cell type group information.
    #        (Default: `True`)
    #    prefix
    #        Prefix of the harmonization columns for all iterations. Default to no prefix.

    #    Returns
    #    ----------
    #    None
    #        A :class:`~pandas.DataFrame` with multiple columns added as the attribute `.reannotation`:
    #        1) **dataset**, datasets where the cells are from.
    #        2) **cell_type**, cell types annotated by the original studies/datasets.
    #        3) **roundN** or **reannotation**, prefixed with `prefix`; cell types reannotated by the harmonization process.
    #        4) **group**, prefixed with `prefix`; annotated cell type groups.
    #    """
    #    if not hasattr(self, 'relation'):
    #        raise AttributeError(
    #                f"🛑 No harmonization result (`.relation`) exists")
    #    reannotation = self.base_distance.cell.set_index('ID', inplace = False, drop = True)
    #    lend = len(self.aligned_datasets)
    #    reannotation[[f"{prefix}round{x}" for x in range(1, lend)]] = UNASSIGN
    #    assignment = self.base_distance.assignment[self.aligned_datasets]
    #    for _, s in self.relation.iterrows():
    #        celltypes = s.values[0::2]
    #        flags = (assignment == celltypes) | np.isin(celltypes, [NOVEL, REMAIN])
    #        non_existing_datasets = self.aligned_datasets[np.isin(celltypes, [NOVEL, REMAIN])]
    #        flags.loc[reannotation.dataset.isin(non_existing_datasets).values, :] = False
    #        for N in range(1, lend):
    #            if (not show_iteration) and (N != lend - 1):
    #                continue
    #            sub_s = s.values[:2*N+1]
    #            sub_celltypes = sub_s[0::2]
    #            if np.all(sub_celltypes == NOVEL) or np.all(sub_celltypes == REMAIN):
    #                continue
    #            reannotation.loc[np.all(flags[self.aligned_datasets[:N+1]], axis = 1).values, f"{prefix}round{N}"] = ' '.join(sub_s)
    #    if add_group:
    #        groups, new_relation = _identify_relation_groups(self.relation, order_row = False, order_column = False)
    #        group_mapping = dict()
    #        for i in range(len(groups)):
    #            group_mapping.update({j: groups[i] for j in new_relation.iloc[i].values})
    #        reannotation[f"{prefix}group"] = (reannotation.dataset + SEP1 + reannotation.cell_type).replace(group_mapping)
    #        reannotation.loc[reannotation[f"{prefix}group"].str.contains(SEP1), f"{prefix}group"] = UNASSIGN
    #    if not show_iteration:
    #        reannotation.rename(columns = {f"{prefix}round{lend-1}": f"{prefix}reannotation"}, inplace = True)
    #        if lend >= 3:
    #            reannotation.drop(columns = [f"{prefix}round{x}" for x in range(1, lend-1)], inplace = True)
    #    self.reannotation = reannotation

    def reannotate(self, prefix: str = '') -> None:
        """
        Reannotate each cell into the harmonized cell type.

        Parameters
        ----------
        prefix
            Prefix of the harmonization columns. Default to no prefix.

        Returns
        ----------
        None
            A :class:`~pandas.DataFrame` with multiple columns added as the attribute `.reannotation`:
            1) **dataset**, datasets where the cells are from.
            2) **cell_type**, cell types annotated by the original studies/datasets.
            3) **reannotation**, prefixed with `prefix`; cell types reannotated by the harmonization process.
            4) **group**, prefixed with `prefix`; annotated cell type groups.
        """
        if not hasattr(self, 'relation'):
            raise AttributeError(
                    f"🛑 No harmonization result (`.relation`) exists")
        reannotation = self.base_distance.cell.set_index('ID', inplace = False, drop = True)
        reannotation[f"{prefix}reannotation"] = UNASSIGN
        assignment = self.base_distance.assignment[self.aligned_datasets]
        for dataset in self.aligned_datasets:
            for celltype in np.setdiff1d(np.unique(self.relation[dataset]), [NOVEL, REMAIN]):
                flag = ((reannotation.dataset == dataset) & (reannotation.cell_type == celltype)).values
                sub_relation = self.relation[self.relation[dataset] == celltype]
                if sub_relation.shape[0] == 1:
                    reannotation.loc[flag, f"{prefix}reannotation"] = ' '.join(sub_relation.iloc[0].values)
                else:
                    sub_assignment = assignment[flag]
                    sums = pd.DataFrame()
                    for _, s in sub_relation.iterrows():
                        celltypes = s.values[0::2]
                        per_s_flag = (sub_assignment == celltypes) | np.isin(celltypes, [NOVEL, REMAIN])
                        sums[' '.join(s.values)] = per_s_flag.sum(axis = 1).values
                    reannotation.loc[flag, f"{prefix}reannotation"] = sums.idxmax(axis = 1).values
        groups, new_relation = _identify_relation_groups(self.relation, order_row = False, order_column = False)
        group_mapping = dict()
        for i in range(len(groups)):
            group_mapping.update({j: groups[i] for j in new_relation.iloc[i].values})
        reannotation[f"{prefix}group"] = (reannotation.dataset + SEP1 + reannotation.cell_type).replace(group_mapping)
        reannotation.loc[reannotation[f"{prefix}group"].str.contains(SEP1), f"{prefix}group"] = UNASSIGN
        self.reannotation = reannotation

    @staticmethod
    def load(alignment_file: str):
        """Load the DistanceAlignment file."""
        with open(alignment_file, "rb") as fh:
            return pickle.load(fh)

    def write(self, file: str) -> None:
        """Write out the DistanceAlignment."""
        file = os.path.splitext(file)[0] + '.pkl'
        with open(file, 'wb') as output:
            pickle.dump(self, output)
